"""TROC Backend.

Troc Authentication using RNC algorithm.
"""
import jwt
import logging
from aiohttp import web
import rapidjson
from navigator.libs.cypher import Cipher
from datetime import datetime, timedelta
from navigator.exceptions import (
    NavException,
    UserDoesntExists,
    InvalidAuth
)
from navigator.conf import (
    PARTNER_KEY,
    CYPHER_TYPE,
    SESSION_TIMEOUT,
    SECRET_KEY,
    AUTH_SESSION_OBJECT
)
import hashlib
import base64
import secrets
from .base import BaseAuthBackend
from .basic import BasicUser

# TODO: add expiration logic when read the token
CIPHER = Cipher(PARTNER_KEY, type=CYPHER_TYPE)


class TrocToken(BaseAuthBackend):
    """TROC authentication Header."""

    user_attribute: str = "user"
    username_attribute: str = "email"

    def __init__(
        self,
        user_attribute: str = None,
        userid_attribute: str = None,
        password_attribute: str = None,
        credentials_required: bool = False,
        authorization_backends: tuple = (),
        **kwargs,
    ):
        super().__init__(
            user_attribute,
            userid_attribute,
            password_attribute,
            credentials_required,
            authorization_backends,
            **kwargs,
        )
        # forcing to use Email as Username Attribute
        self.username_attribute = "email"

    async def validate_user(self, login: str = None):
        # get the user based on Model
        search = {self.username_attribute: login}
        try:
            user = await self.get_user(**search)
            return user
        except UserDoesntExists as err:
            raise UserDoesntExists(
                f"User {login} doesn't exists"
            )
        except Exception as err:
            logging.exception(err)
            raise

    async def get_payload(self, request):
        troctoken = None
        try:
            if "Authorization" in request.headers:
                try:
                    scheme, token = (
                        request.headers.get("Authorization").strip().split(" ")
                    )
                except ValueError:
                    raise web.HTTPForbidden(
                        reason="Invalid authorization Header",
                    )
                if scheme != self.scheme:
                    raise web.HTTPForbidden(
                        reason="Invalid Session scheme",
                    )
            else:
                try:
                    token = request.query.get("auth", None)
                except Exception as e:
                    print(e)
                    return None
        except Exception as err:
            logging.exception(f"TrocAuth: Error getting payload: {err}")
            return None
        return token

    async def authenticate(self, request):
        """ Authenticate, refresh or return the user credentials."""
        try:
            token = await self.get_payload(request)
        except Exception as err:
            raise NavException(
                err, state=400
            )
        if not token:
            raise InvalidAuth(
                "Missing Credentials",
                state=401
            )
        else:
            # getting user information
            # TODO: making the validation of token and expiration
            try:
                data = rapidjson.loads(CIPHER.decode(passphrase=token))
                logging.debug(f'User data: {data!r}')
            except Exception as err:
                raise InvalidAuth(
                    f"Invalid Token: {err!s}", state=401
                )
            # making validation
            try:
                username = data[self.username_attribute]
            except KeyError as err:
                raise InvalidAuth(
                    f"Missing Email attribute: {err!s}", state=401
                )
            try:
                user = await self.validate_user(login=username)
            except UserDoesntExists as err:
                raise UserDoesntExists(err)
            except Exception as err:
                raise NavException(err, state=500)
            try:
                userdata = self.get_userdata(user)
                try:
                    # merging both session objects
                    userdata[AUTH_SESSION_OBJECT] = {
                        **userdata[AUTH_SESSION_OBJECT], **data
                    }
                except Exception as err:
                    logging.exception(err)
                id = user[self.username_attribute]
                username = user[self.username_attribute]
                userdata[self.session_key_property] = id
                usr = BasicUser(data=userdata[AUTH_SESSION_OBJECT])
                usr.id = id
                usr.set(self.username_attribute, username)
                logging.debug(f'User Created > {usr}')
                payload = {
                    self.user_property: user[self.userid_attribute],
                    self.username_attribute: username,
                    "user_id": user[self.userid_attribute],
                }
                token = self.create_jwt(data=payload)
                usr.access_token = token
                # saving user-data into request:
                await self.remember(
                    request, id, userdata, usr
                )
                return {
                    "token": token,
                    **userdata
                }
            except Exception as err:
                logging.exception(f'DjangoAuth: Authentication Error: {err}')
                return False

    async def check_credentials(self, request):
        """ Authentication and create a session."""
        return True



import asyncio
import logging


async def shutdown(loop, signal=None):
    """Cleanup tasks tied to the service's shutdown."""
    if signal:
        print(f"Received exit signal {signal.name}...")
    try:
        tasks = [
            t
            for t in asyncio.all_tasks()
            if t is not asyncio.current_task() and not t.done()
        ]
        [task.cancel() for task in tasks]
        # asyncio.gather(*asyncio.Task.all_tasks()).cancel()
        print(f"Cancelling {len(tasks)} outstanding tasks")
        await asyncio.gather(*tasks, return_exceptions=True)
    except asyncio.CancelledError:
        pass  # Nothing to do or see here
    except Exception as err:
        print("Generic Error", err)


def nav_exception_handler(loop, context):
    """Exception Handler for Asyncio Loops."""
    # first, handle with default handler
    if context:
        loop.default_exception_handler(context)
        exception = context.get("exception")
        # print(exception)
        # print(context)
        try:
            msg = context.get("exception", context["message"])
            logging.error(f"Caught Exception: {msg}")
        except (TypeError, AttributeError, IndexError):
            logging.error(
                "Caught Exception: {}, Context: {}".format(str(exception), str(context))
            )
        # Canceling pending tasks and stopping the loop
        try:
            # loop.run_until_complete(shutdown(loop))
            asyncio.create_task(shutdown(loop))
        except Exception as e:
            print("Shutdown Error: ", e)



"""
Abstract Class for Navigator Authorization Middlewares.
"""

"""
Simply Token Authorization with Callback Support.
"""
import fnmatch
import asyncio
from abc import ABC, abstractmethod
from aiohttp import web, hdrs
from aiohttp.web_urldispatcher import SystemRoute
from functools import partial
from typing import (
    List,
    Dict,
    Callable,
    Tuple,
    Optional,
    Coroutine,
    Any,
    Awaitable
)
from navigator.conf import (
    config,
    CREDENTIALS_REQUIRED,
    SESSION_USER_PROPERTY
)


class base_middleware(ABC):

    anonymous_routes: List = ["/login", "logout", "/static/", "/signin", "/signout", "/_debugtoolbar/"]
    check_static: bool = True
    exclude_routes: Tuple = tuple()
    protected_routes: Tuple = tuple() # list of paths to be protected by middleware
    user_property: str = SESSION_USER_PROPERTY

    def __call__(
            self,
            request: web.Request,
            handler: Callable[[web.Request], Awaitable[web.Response]]
        ):
        """
        Base middleware returns a Awaitable Middleware.
        """
        return self.middleware(request, handler)

    def static_routes(self, request: web.Request):
        """
        Routes declares Statics.
        """
        for r in self.anonymous_routes:
            if request.path.startswith(r):
                return True
        return False

    def excluding_routes(self, request: web.Request):
        for path in self.exclude_routes:
            if fnmatch.fnmatch(request.path, path):
                return True
        return False

    def path_protected(self, request: web.Request):
        if request.path in self.protected_routes:
            return True
        return False


    async def valid_routes(self, request):
        """
        Avoid Authorization on System Routes.
        """
        request.user = None
        if isinstance(request.match_info.route, SystemRoute):  # eg. 404
            return True
        # avoid authorization on exclude list
        if self.exclude_routes and self.excluding_routes(request):
            return True
        if self.check_static is True:
            return self.static_routes(request)
        if request.method == hdrs.METH_OPTIONS:
            return True
        return False

    def get_authorization_header(
            self,
            request: web.Request,
            scheme: Optional[str] = None
        ):
        """
        Get the token and authorization header scheme.
        """
        _scheme = None
        token = None
        if 'Authorization' in request.headers:
            try:
                _scheme, token = request.headers['Authorization'].strip().split(' ')
            except KeyError:
                raise web.HTTPUnauthorized(
                    reason='Token Auth: Missing authorization header',
                )
            except ValueError:
                raise web.HTTPForbidden(
                    reason='Token Auth: Invalid authorization header',
            )
            if scheme is not None and scheme != _scheme:
                raise web.HTTPUnauthorized(
                    reason="Invalid Authorization Scheme"
                )
        else:
            try:
                token = request.query.get("auth", request.headers.get("X-Token", None))
            except KeyError as err:
                token = None
        return [token, _scheme]

    @abstractmethod
    async def middleware(
            self,
            app: web.Application,
            handler: Callable[[web.Request], Awaitable[web.Response]]
        ):
        """
        Abstract Method for declaring Middleware Function.
        """
        pass



#!/usr/bin/env python3
import asyncio
from abc import ABC, abstractmethod
import importlib
import inspect
import asyncio
import uvloop
from pathlib import Path
from typing import Any, Callable, Dict, List
import aiohttp_cors
from navigator.templating import TemplateParser

import aiohttp
from aiohttp import web
from aiohttp.abc import AbstractView

from navigator.conf import (
    APP_NAME,
    APP_DIR,
    DEBUG,
    STATIC_DIR,
    SESSION_TIMEOUT,
    default_dsn
)
from navigator.connections import PostgresPool
# make a home and a ping class
from navigator.resources import home, ping
from navigator.functions import cPrint
# get the authentication library
from navigator.auth import AuthHandler
from navconfig.logging import logging


asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

#######################
##
## PATH CONFIGURATION
##
#######################


class path(object):
    """
    path.
    description: django-like URL router configuration
    """

    method = ""
    url = ""
    handler = None
    name = ""

    def __init__(self, method, url, handler, name=""):
        self.method = method
        self.url = url
        self.handler = handler
        self.name = name


#######################
##
## APPS CONFIGURATION
##
#######################


def app_startup(app_list: list, app: web.Application, context: dict, **kwargs: dict):
    """ Initialize all Apps in the existing Installation."""
    for app_name in app_list:
        obj = None
        try:
            name = app_name.split(".")[1]
            app_class = importlib.import_module(app_name, package="apps")
            obj = getattr(app_class, name)
            instance = obj(context, **kwargs)
            domain = getattr(instance, "domain", None)
            sub_app = instance.App
            if domain:
                app.add_domain(domain, sub_app)
            else:
                app.add_subapp("/{}/".format(name), sub_app)
            # TODO: build automatic documentation
            # add other elements:
            try:
                sub_app['template'] = app["template"]
                # redis connection
                sub_app['redis'] = app['redis']
                if 'database' in app:
                    sub_app['database'] = app['database']
            except Exception as err:
                logging.warning(err)
        except ImportError as err:
            print(err)
            continue


class AppHandler(ABC):
    """
    AppHandler.

    Main Class for registration from Main aiohttp App Creation.
    can register Callbacks, Signals, Route Initialization, etc
     * TODO: adding support for middlewares
     * TODO: get APP names
    """
    _middleware: List = []
    auto_home: bool = True
    enable_notify: bool = False
    enable_static: bool = True
    enable_swagger: bool = True
    auto_doc: bool = False
    enable_auth: bool = True
    staticdir: str = None

    def __init__(
        self,
        context: dict,
        app_name: str = None,
        *args,
        **kwargs
    ) -> None:
        # App Name
        if not app_name:
            self._name = type(self).__name__
        else:
            self._name = app_name
        self.debug = DEBUG
        if not self.staticdir:
            self.staticdir = STATIC_DIR
        self.logger = logging.getLogger(APP_NAME)
        # configuring asyncio loop
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)
        self.app = self.CreateApp()
        # config
        self.app["config"] = context
        # register signals for startup cleanup and shutdown
        self.app.on_startup.append(self.on_startup)
        self.app.on_cleanup.append(self.pre_cleanup)
        self.app.on_cleanup.append(self.on_cleanup)
        self.app.on_shutdown.append(self.on_shutdown)
        self.app.on_response_prepare.append(self.on_prepare)
        # TODO: making automatic discovery of routes
        if self.auto_home:
            self.app.router.add_route("GET", "/", home)

    def get_event_loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    def CreateApp(self) -> web.Application:
        if DEBUG:
            cPrint(f"SETUP APPLICATION: {self._name!s}", level="SUCCESS")
        middlewares = {}
        self.cors = None
        app = web.Application(
            logger=self.logger,
            client_max_size=(1024 * 1024) * 1024
        )
        app.router.add_route("GET", "/ping", ping, name="ping")
        app.router.add_get("/", home, name="home")
        app["name"] = self._name
        # Setup Authentication:
        if self.enable_auth is True:
            self._auth = AuthHandler(
                session_timeout=SESSION_TIMEOUT
            )
            # configuring authentication endpoints
            self._auth.configure(
                app=app,
                handler=self
            )
            app["auth"] = self._auth
            ## add the authorization endpoint endpoint:
            app.router.add_get(
                '/{program}/authorize', self.app_authorization
            )
        # add the other middlewares:
        try:
            for middleware in self._middleware:
                app.middlewares.append(middleware)
        except (ValueError, TypeError):
            pass
        # CORS
        self.cors = aiohttp_cors.setup(
            app,
            defaults={
                "*": aiohttp_cors.ResourceOptions(
                    allow_credentials=True,
                    expose_headers="*",
                    allow_methods="*",
                    allow_headers="*",
                    max_age=3600,
                )
            },
        )
        return app

    @property
    def App(self) -> web.Application:
        return self.app

    @property
    def Name(self) -> str:
        return self._name

    def configure(self) -> None:
        """
        configure.
            making configuration of routes
        """
        if self.enable_static is True:
            # adding statics
            self.app.router.add_static(
                "/static/",
                path=self.staticdir,
                name='static',
                append_version=True,
                show_index=True,
                follow_symlinks=True
            )

    def setup_docs(self) -> None:
        """
        set_cors.
        description: define CORS configuration
        """
        # Configure CORS, swagger and documentation from all routes.
        if self.auto_doc is True:
            for route in list(self.app.router.routes()):
                fn = route.handler
                signature = inspect.signature(fn)
                doc = fn.__doc__
                if doc is None and "OPTIONS" not in route.method:
                    # TODO: making more efficiently
                    fnName = fn.__name__
                    if fnName in ["_handle", "channel_handler", "WebSocket"]:
                        continue
                    if signature.return_annotation:
                        response = str(signature.return_annotation)
                    else:
                        response = aiohttp.web_response.Response
                    doc = """
                    summary: {fnName}
                    description: Auto-Doc for Function {fnName}
                    tags:
                    - Utilities
                    produces:
                    - application/json
                    responses:
                        "200":
                            description: Successful operation
                            content:
                                {response}
                        "404":
                            description: Not found
                        "405":
                            description: invalid HTTP Method
                    """.format(
                        fnName=fnName, response=response
                    )
                    try:
                        fn.__doc__ = doc
                    except (AttributeError, ValueError):
                        pass

    def setup_cors(self, cors):
        for route in list(self.app.router.routes()):
            try:
                if inspect.isclass(route.handler) and issubclass(
                    route.handler, AbstractView
                ):
                    cors.add(route, webview=True)
                else:
                    cors.add(route)
            except (Exception, ValueError) as err:
                # logging.warning(f"Warning on Adding CORS: {err!r}")
                pass
            
    async def app_authorization(self, request: web.Request) -> web.Response:
        app = request.app
        try:
            program = request.match_info['program']
        except Exception as err:
            print(err)
            program = None
        authorization = {
            "status": "Tenant Authorized",
            "program": program
        }
        return web.json_response(authorization, status=200)

    async def on_prepare(self, request, response):
        """
        on_prepare.
        description: Signal for customize the response while is prepared.
        """
        pass

    async def pre_cleanup(self, app):
        """
        pre_cleanup.
        description: Signal for customize the response when server is closing
        """
        pass

    async def on_cleanup(self, app):
        """
        on_cleanup.
        description: Signal for customize the response when server is closing
        """
        pass

    async def on_startup(self, app):
        """
        on_startup.
        description: Signal for customize the response when server is started
        """
        pass

    async def on_shutdown(self, app):
        """
        on_shutdown.
        description: Signal for customize the response when server is shutting down
        """
        pass


class AppBase(AppHandler):
    _middleware = None


class AppConfig(AppHandler):
    """
    AppConfig.

    Class for Configuration of aiohttp SubApps
    """

    template: str = "templates"
    path: Path = None
    _middleware = None
    domain: str = ""
    version: str = '0.0.1'
    description: str = ''
    enable_pgpool: bool = False
    _listener: Callable = None

    def __init__(
        self,
        *args,
        **kwargs
    ):
        self._name = type(self).__name__
        super(AppConfig, self).__init__(*args, **kwargs)
        self.path = APP_DIR.joinpath(self._name)
        # configure templating:
        # TODO: Using the Template Handler exactly like others.
        if self.template:
            try:
                template_dir = self.path.resolve().joinpath(self.template)
                if template_dir.exists():
                    self.app['template'] = TemplateParser(
                        directory=template_dir
                    )
            except Exception as err:
                logging.warning(
                    f'Error Loading Template Parser for SubApp {self._name}: {err}'
                )
        # set the setup_routes
        self.setup_routes()
        # setup swagger
        if self.enable_swagger is True:
            from aiohttp_swagger import setup_swagger
            setup_swagger(
                self.app,
                api_base_url=f"/{self._name}",
                title=f"{self._name} API",
                api_version=self.version,
                description=self.description,
                swagger_url=f"/api/v1/doc",
                ui_version=3,
            )
        self.app.router.add_get(
            '/authorize', self.app_authorization
        )

    def listener(conn, pid, channel, payload, *args):
        print("Notification from {}: {}, {}".format(channel, payload, args))
        
    async def create_connection(self, app, dsn: str = None):
        if not dsn:
            dsn = default_dsn
        pool = PostgresPool(
            dsn=dsn,
            name=f"NAV-{self._name!s}",
            loop=self._loop
        )
        try:
            await pool.startup(app=app)
            app["database"] = pool.connection()
        except Exception as err:
            print(err)
            raise Exception(err)
        if self.enable_notify is True:
            await self.open_connection(app, self._listener)
        return pool.connection()

    async def close_connection(self, app):
        try:
            if self.enable_notify is True:
                conn = app["connection"]
                if conn:
                    await conn.engine().remove_listener(self._name, self.listener)
                    await asyncio.sleep(1)
                await conn.close()
        except Exception as err:
            logging.error("Error closing Interface connection {}".format(err))

    async def open_connection(self, app: web.Application, listener: Callable = None):
        if not listener:
            listener = self.listener
        try:
            conn = await app["database"].acquire()
            app["connection"] = conn
            if conn:
                connection = conn.engine()
                await connection.add_listener(self._name, listener)
                await connection.execute(
                    "NOTIFY \"{}\", '= Starting Navigator Notify System = '".format(
                        self._name
                    )
                )
        except Exception as err:
            raise Exception(err)

    async def on_startup(self, app):
        # enabled pgpool
        if self.enable_pgpool is True:
            try:
                await self.create_connection(app, default_dsn)
            except Exception:
                pass

    async def on_shutdown(self, app):
        if self.enable_pgpool is True:
            try:
                await self.close_connection(app)
            except Exception:
                pass

    def setup_routes(self):
        """Setup Routes (URLS) pointing to paths on AppConfig."""
        # set the urls
        # TODO: automatic module loader
        try:
            cls = importlib.import_module(
                "{}.{}".format("apps.{}".format(self._name), "urls"), package="apps"
            )
            routes = getattr(cls, "urls")
        except ImportError as err:
            print(err)
            return False
        for route in routes:
            # print(route, route.method)
            if inspect.isclass(route.handler) and issubclass(
                route.handler, AbstractView
            ):
                if route.method is None:
                    r = self.app.router.add_view(
                        route.url, route.handler, name=route.name
                    )
                    self.cors.add(r, webview=True)
                elif not route.method:
                    r = self.app.router.add_view(
                        route.url, route.handler, name=route.name
                    )
                elif route.method == "*":
                    r = self.app.router.add_route(
                        "*", route.url, route.handler, name=route.name
                    )
                else:
                    if route.method == "get":
                        r = self.app.router.add_get(
                            route.url, route.handler, name=route.name
                        )
                    elif route.method == "post":
                        r = self.app.router.add_post(
                            route.url, route.handler, name=route.name
                        )
                    elif route.method == "delete":
                        r = self.app.router.add_delete(
                            route.url, route.handler, name=route.name
                        )
                    elif route.method == "patch":
                        r = self.app.router.add_patch(
                            route.url, route.handler, name=route.name
                        )
                    elif route.method == "put":
                        r = self.app.router.add_put(
                            route.url, route.handler, name=route.name
                        )
                    else:
                        raise (
                            "Unsupported Method for Route {}, program: {}".format(
                                route.method, self._name
                            )
                        )
                        return False
                    self.cors.add(r, webview=True)
            elif inspect.isclass(route.handler):
                r = self.app.router.add_view(route.url, route.handler, name=route.name)
                self.cors.add(r, webview=True)
            else:
                if not route.method:
                    r = self.app.router.add_route(
                        "*", route.url, route.handler, name=route.name
                    )
                else:
                    if route.method == "get":
                        r = self.app.router.add_get(
                            route.url, route.handler, name=route.name, allow_head=False
                        )
                    elif route.method == "post":
                        r = self.app.router.add_post(
                            route.url, route.handler, name=route.name
                        )
                    elif route.method == "delete":
                        r = self.app.router.add_delete(
                            route.url, route.handler, name=route.name
                        )
                    elif route.method == "patch":
                        r = self.app.router.add_patch(
                            route.url, route.handler, name=route.name
                        )
                    elif route.method == "put":
                        r = self.app.router.add_put(
                            route.url, route.handler, name=route.name
                        )
                    else:
                        raise (
                            "Unsupported Method for Route {}, program: {}".format(
                                route.method, self._name
                            )
                        )
                        return False
                    self.cors.add(r)

    async def app_authorization(self, request: web.Request) -> web.Response:
        app = request.app
        try:
            program = request.match_info['program']
        except Exception as err:
            program = self.__class__.__name__
        authorization = {
            "status": "Tenant Authorized",
            "program": program
        }
        return web.json_response(authorization, status=200)



"""OktaAuth.

Description: Backend Authentication/Authorization using Okta Service.
"""
import logging
from aiohttp import web
from .oauth import OauthAuth
import requests
from navigator.exceptions import (
    NavException
)
from navigator.conf import (
    GITHUB_CLIENT_ID,
    GITHUB_CLIENT_SECRET
)


class GithubAuth(OauthAuth):
    """GithubAuth.

    Description: Authentication Backend using Third-party GitHub Service.
    """
    userid_attribute: str = "login"
    user_attribute: str = "name"
    username_attribute: str = "email"
    _service_name: str = "github"

    def configure(self, app, router, handler):
        super(GithubAuth, self).configure(app, router, handler) # first, configure parents
        
        # auth paths.
        self.base_url = 'https://api.github.com/'
        self.authorize_uri = 'https://github.com/login/oauth/authorize'
        self.userinfo_uri = 'https://api.github.com/user'
        self._issuer = 'https://api.github.com/'
        self._token_uri = 'https://github.com/login/oauth/access_token'

    
    async def get_credentials(self, request: web.Request):
        qs = {
            "client_id": f"{GITHUB_CLIENT_ID}",
            "client_secret": GITHUB_CLIENT_SECRET,
            "scope": "user:email"
        }
        return qs

    async def auth_callback(self, request: web.Request):
        try:
            auth_response = self.get_auth_response(request)
            code = self.get_auth_code(auth_response)
        except Exception as err:
            message = f"Github: Missing Auth Token: {auth_response}"
            logging.exception(message)
            response = {
                "message": message
            }
            return web.json_response(response, status=403)
        grant = {
            "client_id": f"{GITHUB_CLIENT_ID}",
            "client_secret": GITHUB_CLIENT_SECRET,
            "code": code,
            'scope': "user:email"
        }
        result = await self.post(self._token_uri, json=grant)
        if 'error' in result:
            message = f"Github Error getting Access Token: {result}"
            logging.exception(message)
            response = {
                "message": message
            }
            return web.json_response(response, status=403)
        else:
            access_token = result['access_token']
            token_type = result['token_type']
            # then, will get user info:
            try:
                headers = {
                    "Accept": "application/vnd.github.v3+json",
                    "X-OAuth-Scopes": "repo, user"
                }
                data = await self.get(self.userinfo_uri, token=access_token, token_type='token', headers=headers)
                if data:
                    userdata, uid = self.build_user_info(data)
                    # also, user information:
                    data = await self.create_user(request, uid, userdata, access_token)
                    # Redirect User to HOME
                    return self.home_redirect(request, token=data["token"], token_type='Bearer')
                else:
                    return self.redirect(uri=self.login_failed_uri)
            except Exception as err:
                logging.exception(err)
                return self.redirect(uri=self.login_failed_uri)


    async def logout(self, request):
        pass

    async def finish_logout(self, request):
        pass

    async def check_credentials(self, request):
        """ Authentication and create a session."""
        return True



import os
import sys
import json
import asyncio
import logging
from pathlib import Path

from aiofile import AIOFile

from . import BaseCommand, CommandError

logger = logging.getLogger('navigator')
loop = asyncio.get_event_loop()


def read_file(path, filename):
    f = open(path.joinpath("navigator", "commands", "templates", filename), "r")
    return f.read()


def create_dir(dir, name, touch_init: bool = False):
    try:
        path = dir.joinpath(name)
        path.mkdir(parents=True, exist_ok=True)
        if touch_init is True:
            # create a __init__ file
            save_file(path, "__init__.py", content="#!/usr/bin/env python3")
    except FileExistsError as exc:
        pass


def delete_file(dir, name):
    path = dir.joinpath(name)
    try:
        path.unlink(missing_ok=True)
        return True
    except Exception as err:
        print(err)
        return False


def save_file(dir, filename, content):
    async def main(filename, content):
        try:
            path = dir.joinpath(filename)
            async with AIOFile(path, "w+") as afp:
                await afp.write(content)
                await afp.fsync()
            return True
        except Exception as err:
            print(err)
            logging.error(err)
            return False

    return loop.run_until_complete(main(filename, content))


def drive_permission(client: str, secret: str, project: str = "navigator") -> str:
    permission = {
        "web": {
            "client_id": f"{client!s}.apps.googleusercontent.com",
            "project_id": project,
            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
            "token_uri": "https://oauth2.googleapis.com/token",
            "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
            "client_secret": secret,
            "redirect_uris": ["http://localhost:8090/"],
            "javascript_origins": ["http://localhost:8090"],
        }
    }
    return json.dumps(permission)


class EnvCommand(BaseCommand):
    help = "Enviroment Commands for Navigator"

    def parse_arguments(self, parser):
        parser.add_argument("--enable-notify", type=bool)
        parser.add_argument("--process-services", type=bool)
        parser.add_argument("--file_env", type=str)
        parser.add_argument("--client", type=str)
        parser.add_argument("--secret", type=str)
        parser.add_argument("--project", type=str)

    def create(self, options, **kwargs):
        """
        Create can used to create a new Enviroment from scratch
        """
        path = Path(kwargs["project_path"]).resolve()
        ini = read_file(path, "ini.tpl")
        settings = read_file(path, "settings.tpl")
        localsettings = read_file(path, "localsettings.tpl")
        run = read_file(path, "run.tpl")
        app = read_file(path, "app.tpl")

        output = "Enviroment Done."
        if options.debug:
            self.write(":: Creating a New Navigator Enviroment", level="INFO")
            self.write("= wait a few minutes", level="WARN")
        # apps, etc, env, services, settings, static/images/js/css, templates
        self.write("* First Step: Creating Directory structure")
        create_dir(path, "apps", touch_init=True)
        create_dir(path, "env/testing")
        create_dir(path, "etc")
        # create_dir(path, "log")
        create_dir(path, "services", touch_init=True)
        create_dir(path, "resources", touch_init=True)
        create_dir(path, "settings", touch_init=True)
        create_dir(path, "static/images")
        create_dir(path, "static/js")
        create_dir(path, "static/css")
        create_dir(path, "templates")
        self.write("* Second Step: Creation of Empty .env File")
        # save_file(path, "env/.env", env)
        # save_file(path, "env/testing/.env", env)
        save_file(path, "etc/navigator.ini", ini)
        # TODO: download from Google Drive, if possible
        self.write("* Third Step: Creation of Empty settings.py File")
        save_file(path, "settings/settings.py", settings)
        save_file(path, "settings/local_settings.py.example", localsettings)
        self.write("* Fourt Step: Creation of a run.py File")
        save_file(path, "run.py", run)
        save_file(path, "app.py", app)
        self.write("* Fifth Step: adding a *home* template")
        home = read_file(path, "home.html")
        css = read_file(path, "styles.css")
        js = read_file(path, "scripts.js")
        save_file(path, "templates/home.html", home)
        save_file(path, "static/css/styles.css", css)
        save_file(path, "static/js/scripts.js", js)
        return output

    def get_env(self, options, **kwargs):
        """get_env.

        Getting a new environment file from Google Drive.
        """
        path = Path(kwargs["project_path"]).resolve()
        file_env = options.file_env
        # first: removing existing credentials
        delete_file(path, "env/credentials.txt")
        # saving the credentials into a new file
        save_file(path, "env/file_env", file_env)
        # preparing the environment:
        # set SITE_ROOT:
        os.environ["SITE_ROOT"] = str(path)
        print("SITE ROOT: ", os.getenv("SITE_ROOT"))
        # set configuration for navconfig
        os.environ["NAVCONFIG_ENV"] = "drive"
        os.environ["NAVCONFIG_DRIVE_CLIENT"] = "env/credentials.txt"
        os.environ["NAVCONFIG_DRIVE_ID"] = file_env
        # get the drive permission
        client = options.client
        secret = options.secret
        project = options.project
        content = drive_permission(client, secret, project)
        save_file(path, "client_secrets.json", content)
        from navconfig import config

        config.save_environment("drive")
        return "Done."



"""
APIKEY infraestructure for NAVIGATOR.

Simple API Key/Secret Validator for Navigator using a Middleware.
"""
import logging
from aiohttp import web
from typing import (
    Optional,
    Coroutine,
    Tuple
)
from .abstract import base_middleware


class apikey_middleware(base_middleware):
    def __init__(
        self,
        user_fn: Optional[Coroutine] = None,
        protected_routes: Optional[Tuple] = tuple()
    ):
        """
        Check if an Auth Token was provided and returns based on
        an user Callback Function.

        The Optional Callback can receive the Payload of the deciphered token
        and the Request.
        """
        if user_fn is not None and not callable(user_fn):
            raise RuntimeError(
                f"If defined, User Function {user_fn!s} need to be Callable."
            )
        self._fn = user_fn
        if protected_routes:
            self.protected_routes = protected_routes

    def get_authorization_header(
            self,
            request: web.Request,
            scheme: Optional[str] = None
        ):
        """
        Get the key and secret from header
        """
        token = None
        if 'x-api-key' in request.headers:
            try:
                token = request.headers['x-api-key'].strip()
            except KeyError:
                raise web.HTTPUnauthorized(
                    reason='API Key Auth: Missing authorization header',
                )
            except ValueError:
                raise web.HTTPForbidden(
                    reason='API Key Auth: Invalid authorization header',
            )
        else:
            try:
                token = request.query.get("api_key").strip()
            except KeyError as err:
                token = None
        return token

    async def middleware(self, app, handler):
        @web.middleware
        async def middleware(request):
            if await self.valid_routes(request):
                return await handler(request)
            try:
                key = self.get_authorization_header(request)
            except Exception as err:
                logging.exception(err)
                key = None
            if self.path_protected(request): # is a protected site.
                if not key:
                    raise web.HTTPForbidden(
                        reason='Missing API Key Authentication',
                    )
                try:
                    db = app['database']
                    async with await db.acquire() as conn:
                        payload = await conn.fetch_one(
                            "SELECT user_id, name from public.api_keys where token = $1 AND (expiration >= extract(epoch from now()) or expiration = 0)",
                            key
                        )
                    if not payload:
                        raise web.HTTPForbidden(
                            reason="Access is Restricted"
                        )
                except Exception as err:
                    raise web.HTTPBadRequest(
                        reason=f"API Key Decryption Error: {err}"
                    )
                try:
                    if self._fn:
                        user = await self._fn(payload, request)
                        if user:
                            request[self.user_property] = user
                            request.user = user
                        else:
                            raise web.HTTPForbidden(
                                reason='Access Restricted'
                            )
                except Exception as err:
                    raise web.HTTPBadRequest(
                        reason=f'Exception on Callable called by {__name__!s} {err!s}'
                    )
            return await handler(request)
        return middleware



"""
Simply Token Authorization with Callback Support.
"""
from aiohttp import web
from typing import (
    Dict,
    Callable,
    Tuple,
    Optional,
    Coroutine,
    Any
)
from .abstract import base_middleware
from navigator.conf import (
    config,
    CREDENTIALS_REQUIRED,
    SESSION_USER_PROPERTY
)


class token_middleware(base_middleware):
    def __init__(
        self,
        user_fn: Coroutine,
        user_property: str = SESSION_USER_PROPERTY,
        exclude_routes: Optional[Tuple] = tuple()
    ):
        """
        Check if an Auth Token was provided and returns based on
        an user Callback Function.

        Args:
            user_fn: any User Callable Coroutine, to process the received token.
              TODO: if not, can we use own token factory.
            user_property: the User Object returned by callback is saved
              on this property under the request.
            exclude_routes: any path that needs to be excluded for Token Auth
        Returns:
            handler if User exists, HTTP Forbidden if callback returns false.
            HTTP Unauthorized if Token is missing (only if credential required is TRUE)
        """
        if not callable(user_fn):
            raise RuntimeError(
                f"User Function {user_fn!s} need to be an Asyncio Coroutine."
            )
            # TODO: adds a generic list of exclude routes
        self._fn = user_fn

        if exclude_routes is None:
            self.exclude_routes = config.get('EXCLUDED_ROUTES', tuple())
        else:
            self.exclude_routes = exclude_routes
        # user property
        self.user_property = user_property

    async def middleware(self, app, handler):
        @web.middleware
        async def middleware(request):
            if await self.valid_routes(request):
                return await handler(request)
            try:
                token, scheme = self.get_authorization_header(request)
            except KeyError as err:
                token = None
            if CREDENTIALS_REQUIRED is True:
                if not token:
                    raise web.HTTPForbidden(
                        reason='Token Auth: Invalid authorization Token',
                    )
                try:
                    user = await self._fn(token, scheme, request)
                    if user:
                        request[self.user_property] = user
                        request.user = user
                        print(user)
                    else:
                        raise web.HTTPForbidden(
                            reason='Access Restricted'
                        )
                except Exception as err:
                    raise web.HTTPBadRequest(
                        reason=f'Token Auth: Exception on Callable Return {err!s}'
                    )
            return await handler(request)
        return middleware



#!/usr/bin/env python
import importlib
import os
import sys
import traceback
from argparse import SUPPRESS, ArgumentParser, HelpFormatter
from inspect import signature
from io import TextIOBase
from typing import Any, Callable, Dict, List

from navigator import get_version
from navigator.conf import INSTALLED_APPS
from navigator.functions import cPrint


class CommandError(Exception):
    """
    Exception Base Class for raise problems in the execution of a Command
    """

    pass


class CommandNotFound(Exception):
    """
    Exception Base Class for raise problems in the execution of a Command
    """

    pass


class BaseCommand(object):
    parser: Callable
    args: List = []
    action: str = ""
    help: str = "Base Help Command"
    epilog: str = ""

    def __init__(self, args):
        self.args = args
        self.parser = ArgumentParser(
            description=self.help,
            epilog=self.epilog if self.epilog else self.help,
            add_help=False,
        )
        self.parser.add_argument(
            "-v", "--version", action="version", version="%(prog)s 1.0"
        )
        self.parser.add_argument(
            "-h", "--help", action="help", default=SUPPRESS, help="Display this Message"
        )
        self.parser.add_argument(
            "-d", "--debug", action="store_true", help="Enable Debug"
        )
        self.parser.add_argument(
            "--traceback",
            action="store_true",
            help="Return the Traceback on CommandError",
        )
        # get action:
        self.action = self.args.pop(0)
        self.parse_arguments(self.parser)

    def write(self, message, level="INFO"):
        if message:
            cPrint(message, level=level)

    def parse_arguments(self, parser):
        """
        parse_arguments.
            allow for subclassed comands to add custom arguments
        """
        pass

    def get_version(self):
        """
        get_version
            Return the current Navigator Version
        """
        return "Navigator: v.{}".format(get_version())

    def handle(self, **kwargs):
        output: str = ""
        try:

            # calling the internal function:
            if not hasattr(self, self.action):
                self.write(
                    "Error: Method **{}** not found on {}".format(
                        self.action, str(self)
                    ),
                    level="ERROR",
                )
                raise CommandNotFound(
                    "Method {} from {} not Found".format(self, self.action)
                )
            fn = getattr(self, self.action)
            # adding an epilog using the docstring
            self.parser.epilog = str(fn.__doc__)
            # parsing current arguments
            options = self.parser.parse_args(self.args)
            if options.debug:
                self.write("Executing : {} Command.".format(self.action), level="DEBUG")
            sig = signature(fn)
            try:
                if len(sig.parameters) > 0:
                    # send parameters to method
                    output = fn(options, **kwargs)
                else:
                    output = fn()
            except Exception as err:
                if options.traceback:
                    print(traceback.format_exc())
                raise CommandError(
                    "Error Calling Method: {}, error: {}".format(self.action, err)
                )
        except Exception as err:
            if options.traceback:
                print(traceback.format_exc())
            raise CommandError("Error Parsing arguments: {}".format(err))
        finally:
            self.write(output, level="INFO")
            return output


def get_command(command: str = "troc", clsname: str = "", pathname: str = "navigator"):
    try:
        if pathname:
            classpath = "{path}.commands.{command}".format(
                path=pathname, command=command
            )
        else:
            classpath = "commands.{command}".format(command=command)
        module = importlib.import_module(classpath, package="commands")
        cls = getattr(module, clsname)
        return cls
    except (ModuleNotFoundError, ImportError):
        # last resort: direct commands on source
        raise CommandNotFound(f"Command {clsname} was not found on {pathname}")


def run_command(**kwargs):
    """
    Running a command in Navigator Enviroment

    Command is running in the form:
    manage.py {command} {instructions}
    example: manage.py app create
    """
    if len(sys.argv) > 1:
        args = sys.argv
        script = args.pop(0)
        command = args.pop(0)
        if command is not None:
            # if command is a program, the behavior is different:
            program = "apps.{}".format(command)
            if program in INSTALLED_APPS:
                # is a program command
                cmd = args.pop(0)
                if not args:
                    args.append(cmd)
                clsCommand = "{}Command".format(cmd.capitalize())
                # classpath = "{program}.commands.{provider}".format(
                #     program=program, provider=clsCommand
                # )
                try:
                    cls = get_command(
                        command=clsCommand, clsname=clsCommand, pathname=program
                    )
                    # module = importlib.import_module(classpath, package=clsCommand)
                    # cls = getattr(module, clsCommand)
                except CommandNotFound:
                    raise CommandNotFound(
                        "Command %s for program %s was not found"
                        % (clsCommand, program)
                    )
            else:
                clsCommand = "{}Command".format(command.capitalize())
                # check if is a Navigator Command
                try:
                    cls = get_command(
                        command=command, clsname=clsCommand, pathname="navigator"
                    )
                except CommandNotFound:
                    # last resort: direct commands on source
                    try:
                        cls = get_command(
                            command=command, clsname=clsCommand, pathname=""
                        )
                    except CommandNotFound:
                        raise CommandNotFound(
                            "Command %s was not found o program doesnt exists"
                            % clsCommand
                        )
            try:
                cmd = cls(args)
                output = cmd.handle(**kwargs)
            except Exception as err:
                print(err)



"""Navigator Auth.

Navigator Authentication/Authorization system.

AuthHandler is the Authentication/Authorization system for NAV,
Supporting:
 * multiple authentication backends
 * authorization exceptions via middlewares
 * Session Support (in the top of aiohttp-session)
"""
from textwrap import dedent
import importlib
import logging
from aiohttp import web
from typing import Dict, Iterable
from .authorizations import *
from navigator.functions import json_response

from navigator.auth.sessions.storages import (
    RedisStorage,
)

from navigator.exceptions import (
    NavException,
    UserDoesntExists,
    InvalidAuth,
    FailedAuth
)
from navigator.conf import (
    CREDENTIALS_REQUIRED,
    AUTHENTICATION_BACKENDS,
    AUTHORIZATION_BACKENDS,
    AUTHORIZATION_MIDDLEWARES,
    AUTH_USER_MODEL,
    SESSION_KEY
)
# from navigator.auth.sessions import get_session, new_session

class AuthHandler(object):
    """Authentication Backend for Navigator."""

    _template = """
        <!doctype html>
            <head></head>
            <body>
                <p>{message}</p>
                <form action="/login" method="POST">
                  Login:
                  <input type="text" name="login">
                  Password:
                  <input type="password" name="password">
                  <input type="submit" value="Login">
                </form>
                <a href="/logout">Logout</a>
            </body>
    """
    backends: Dict = None
    _required: bool = False
    _middlewares: Iterable = tuple()

    def __init__(
        self,
        auth_scheme="Bearer",
        **kwargs,
    ):
        self._session = None
        self._template = dedent(self._template)
        authz_backends = self.get_authorization_backends(
            AUTHORIZATION_BACKENDS
        )
        # Get User Model:
        try:
            user_model = self.get_usermodel(AUTH_USER_MODEL)
        except Exception:
            # TODO: more feature content, exceptions.
            raise
        args = {
            "credentials_required": CREDENTIALS_REQUIRED,
            "scheme": auth_scheme,
            "authorization_backends": authz_backends,
            "user_model": user_model,
            **kwargs,
        }
        # get the authentication backends (all of the list)
        self.backends = self.get_backends(**args)
        self._middlewares = self.get_authorization_middlewares(
            AUTHORIZATION_MIDDLEWARES
        )
        # TODO: Session Support with parametrization (other backends):
        self._session = RedisStorage()

    def get_backends(self, **kwargs):
        backends = {}
        for backend in AUTHENTICATION_BACKENDS:
            try:
                parts = backend.split(".")
                bkname = parts[-1]
                classpath = ".".join(parts[:-1])
                module = importlib.import_module(classpath, package=bkname)
                obj = getattr(module, bkname)
                logging.debug(f'Loading Auth Backend {bkname}')
                backends[bkname] = obj(**kwargs)
            except ImportError:
                raise Exception(f"Error loading Auth Backend {backend}")
        return backends

    def get_usermodel(self, model: str, **kwargs):
        try:
            parts = model.split(".")
            name = parts[-1]
            classpath = ".".join(parts[:-1])
            module = importlib.import_module(classpath, package=name)
            obj = getattr(module, name)
            return obj
        except ImportError:
            raise Exception(f"Error loading Auth User Model {model}")

    async def on_startup(self, app):
        """
        Some Authentication backends need to call an Startup.
        """
        for name, backend in self.backends.items():
            try:
                await backend.on_startup(app)
            except Exception as err:
                print(err)
                logging.exception(
                    f"Error on Startup Auth Backend {name} init: {err!s}"
                )

    async def on_cleanup(self, app):
        """
        Cleanup the processes
        """
        for name, backend in self.backends.items():
            try:
                await backend.on_cleanup(app)
            except Exception as err:
                print(err)
                logging.exception(
                    f"Error on Cleanup Auth Backend {name} init: {err!s}"
                )

    def get_authorization_backends(self, backends: Iterable) -> tuple:
        b = []
        for backend in backends:
            # TODO: more automagic logic
            if backend == "hosts":
                b.append(authz_hosts())
            elif backend == "allow_hosts":
                b.append(authz_allow_hosts())
        return b

    def get_authorization_middlewares(self, backends: Iterable) -> tuple:
        b = tuple()
        for backend in backends:
            try:
                parts = backend.split(".")
                bkname = parts[-1]
                classpath = ".".join(parts[:-1])
                module = importlib.import_module(classpath, package=bkname)
                obj = getattr(module, bkname)
                b.append(obj)
            except ImportError:
                raise Exception(
                    f"Error loading Authz Middleware {backend}"
                )
        return b

    async def api_logout(self, request: web.Request) -> web.Response:
        """Logout.
        API-based Logout.
        """
        try:
            await self._session.forgot(request)
            return web.json_response(
                {
                    "message": "Logout successful",
                    "state": 202
                },
                status=202
            )
        except Exception as err:
            print(err)
            raise web.HTTPUnauthorized(
                reason=f"Logout Error {err!s}"
            )

    async def api_login(self, request: web.Request) -> web.Response:
        """Login.

        API based login.
        """
        app = request.app
        # first: getting header for an existing backend
        method = request.headers.get('X-Auth-Method')
        if method:
            try:
                backend = self.backends[method]
            except KeyError:
                raise web.HTTPUnauthorized(
                    text=f"Unacceptable Auth Method {method}",
                    content_type="application/json"
                )
            try:
                userdata = await backend.authenticate(request)
                if not userdata:
                    raise web.HTTPForbidden(
                        reason='User was not authenticated'
                    )
            except FailedAuth as err:
                raise web.HTTPForbidden(
                    reason=f"{err!s}"
                )
            except InvalidAuth as err:
                logging.exception(err)
                raise web.HTTPForbidden(
                    reason=f"{err!s}"
                )
            except UserDoesntExists as err:
                raise web.HTTPForbidden(
                    reason="User Doesn't exists: {err!s}"
                )
            except Exception as err:
                raise web.HTTPClientError(
                    reason=f"{err!s}"
                )
            return json_response(userdata, state=200)
        else:
            # second: if no backend declared, will iterate over all backends
            userdata = None
            for bk, backend in self.backends.items():
                try:
                    # check credentials for all backends
                    userdata = await backend.authenticate(request)
                    if userdata:
                        break
                except (
                    NavException,
                    UserDoesntExists,
                    InvalidAuth,
                    FailedAuth
                ) as err:
                    continue
                except Exception as err:
                    return web.HTTPClientError(
                        reason=err,
                        status=401
                    )
            # if not userdata, then raise an not Authorized
            if not userdata:
                raise web.HTTPForbidden(
                    reason="Login Failure in all Auth Methods."
                )
            else:
                # at now: create the user-session
                try:
                    session = await self._session.new_session(request, userdata)
                except Exception as err:
                    print(err)
                    return web.HTTPUnauthorized(
                        reason=f"Error Creating User Session: {err!s}"
                    )
                return json_response(userdata, state=200)

    # Session Methods:
    async def forgot_session(self, request: web.Request):
        await self._session.forgot(request)

    async def create_session(self, request: web.Request, data: Iterable):
        return await self._session.new_session(request, data)

    async def get_session(self, request: web.Request) -> web.Response:
        """ Get user data from session."""
        session = None
        try:
            session = await self._session.get_session(request)
        except NavException as err:
            response = {
                "message": "Session Error",
                "error": err.message,
                "status": err.state,
            }
            return web.json_response(response, status=err.state)
        except Exception as err:
            return web.HTTPClientError(
                reason=err,
                status=401
            )
        if not session:
            try:
                session = await self._session.get_session(request)
            except Exception as e:
                print(e)
                # always return a null session for user:
                session = await self._session.new_session(request, {})
        userdata = dict(session)
        try:
            del userdata['user']
        except KeyError:
            pass
        return web.json_response(userdata, status=200)

    def configure(
            self,
            app: web.Application,
            handler
        ) -> web.Application:
        # first, add signals:
        # startup operations over authentication backend
        app.on_startup.append(
            self.on_startup
        )
        # cleanup operations over authentication backend
        app.on_cleanup.append(
            self.on_cleanup
        )
        router = app.router
        router.add_route(
            "GET",
            "/api/v1/login",
            self.api_login,
            name="api_login"
        )
        router.add_route(
            "POST",
            "/api/v1/login",
            self.api_login,
            name="api_login_post"
        )
        router.add_route(
            "GET",
            "/api/v1/logout",
            self.api_logout,
            name="api_logout"
        )
        # get the session information for a program (only)
        router.add_route(
            "GET",
            "/api/v1/session/{program}",
            self.get_session,
            name="api_session_tenant",
        )
        # get all user information
        router.add_route(
            "GET",
            "/api/v1/user/session",
            self.get_session,
            name="api_session"
        )
        # the backend add a middleware to the app
        mdl = app.middlewares
        # configuring Session Object
        self._session.configure_session(app)
        # if authentication backend needs initialization
        for name, backend in self.backends.items():
            try:
                backend.configure(app, router, handler)
                if hasattr(backend, "auth_middleware"):
                    # add the middleware for Basic Authentication
                    mdl.append(backend.auth_middleware)
            except Exception as err:
                print(err)
                logging.exception(
                    f"Error on Auth Backend {name} init: {err!s}"
                )
        # at last: add other middleware support
        if self._middlewares:
            for mid in self._middlewares:
                mdl.append(mid)
        return app

async def get_auth(
        request: web.Request
) -> str:
    """
    Get the current User ID from Request
    """
    id = request.get(SESSION_KEY, None)
    if id:
        return id

async def get_userdata(
        request: web.Request
) -> str:
    """
    Get the current User ID from Request
    """
    data = request.get(self.user_property, None)
    if data:
        return data
    else:
        raise web.HTTPForbidden(
            reason="Auth: User Data is missing on Request."
        )



import decimal
import json
import os
import sys
import uuid
from datetime import datetime
from decimal import Decimal
from enum import Enum

import asyncpg


class DateEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return str(obj)
        else:
            return str(object=obj)
        return json.JSONEncoder.default(self, obj)


class IntRangeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, asyncpg.Range):
            up = obj.upper
            if isinstance(obj.upper, int):
                up = obj.upper - 1  # discrete representation
            return [obj.lower, up]
        else:
            return str(object=obj)
        return json.JSONEncoder.default(self, obj)


class DefaultEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, "hex"):
            return obj.hex
        elif isinstance(obj, Enum):
            if not obj.value:
                return None
            else:
                return str(obj.value)
        elif isinstance(obj, uuid.UUID):
            try:
                return str(obj)
            except Exception as e:
                return obj.hex
        elif isinstance(obj, decimal.Decimal):
            return float(obj)
        elif isinstance(obj, Decimal):
            return str(obj)
        elif hasattr(obj, "isoformat"):
            return obj.isoformat()
        elif isinstance(obj, asyncpg.Range):
            return [obj.lower, obj.upper]
        else:
            return str(object=obj)
        return json.JSONEncoder.default(self, obj)



"""User sessions for Navigator and aiohttp.web server."""

__version__ = '0.0.1'

from aiohttp import web
import asyncio
import time
import logging
from .storages import SessionData
from navigator.conf import (
    SESSION_NAME,
    SESSION_PREFIX,
    SESSION_TIMEOUT,
    SESSION_KEY,
    SESSION_STORAGE,
    SESSION_OBJECT
)

async def new_session(request: web.Request, userdata: dict = {}) -> SessionData:
    storage = request.get(SESSION_STORAGE)
    if storage is None:
        raise RuntimeError(
            "Missing Configuration for Session Middleware."
        )
    session = await storage.new_session(request, userdata)
    if not isinstance(session, SessionData):
        raise RuntimeError(
            "Installed {!r} storage should return session instance "
            "on .load_session() call, got {!r}.".format(storage, session))
    request[SESSION_OBJECT] = session
    return session

async def get_session(
        request: web.Request,
        userdata: dict = {},
        new: bool = False
) -> SessionData:
    session = request.get(SESSION_OBJECT)
    logging.debug(f'SESSION IS: {session!s} for object {SESSION_OBJECT!s}')
    if session is None:
        storage = request.get(SESSION_STORAGE)
        logging.debug(f'SESSION STORAGE IS {storage}')
        if storage is None:
            raise RuntimeError(
                "Missing Configuration for Session Middleware."
            )
        # using the storage session for Load an existing Session
        try:
            session = await storage.load_session(
                request=request,
                userdata=userdata,
                new=new
            )
        except Exception as err:
            raise RuntimeError(
                "Error Loading user Session."
            )
        request[SESSION_OBJECT] = session
        request['session'] = session
        if new is True and not isinstance(session, SessionData):
            raise RuntimeError(
                "Installed {!r} storage should return session instance "
                "on .load_session() call, got {!r}.".format(storage, session))
    return session



"""
Model System for Navigator Auth.

Model for User, Group and Roles for Navigator Auth.
"""
from asyncdb.models import Model, Column, Field
from typing import (
    Optional,
    List,
    Dict,
    Union,
    Tuple,
    Any,
    Callable
)
from datetime import datetime
from navigator.conf import default_dsn, USERS_TABLE


class User(Model):
    """Basic User notation."""

    user_id: int = Column(required=False, primary_key=True)
    first_name: str
    last_name: str
    email: str = Column(required=False, max=254)
    password: str = Column(required=False, max=128)
    last_login: datetime = Column(required=False)
    username: str = Column(required=False)
    is_superuser: bool = Column(required=True, default=False)
    is_active: bool = Column(required=True, default=True)
    is_new: bool = Column(required=True, default=True)
    title: str = Column(equired=False, max=90)
    registration_key: str = Column(equired=False, max=512)
    reset_pwd_key: str = Column(equired=False, max=512)
    avatar: str = Column(max=512)

    def __getitem__(self, item):
        return getattr(self, item)

    @property
    def display_name(self):
        return f"{self.first_name} {self.last_name}"

    class Meta:
        driver = "pg"
        dsn: str = default_dsn
        name = USERS_TABLE
        schema = "public"
        strict = True
        frozen = False
        connection = None

# # TODO: add autoincrement feature, read the last id and plus 1
# class Group(Model):
#     id: int = Column(required=True, primary_key=True)
#     name: int = Column(required=True)

#     class Meta:
#         driver = "pg"
#         dsn: str = default_dsn
#         name = "auth_group"
#         schema = "public"
#         strict = True
#         frozen = False
#         connection = None




"""ExternalAuth Backend.

Abstract Model to any Oauth2 or external Auth Support.
"""
import logging
import aiohttp
from aiohttp import web, hdrs
from .base import BaseAuthBackend
from typing import (
    Dict,
    List,
    Any
)
from abc import abstractmethod
from navigator.conf import (
    AUTH_SESSION_OBJECT,
    AUTH_LOGIN_FAILED_URI,
    AUTH_REDIRECT_URI
)
from navigator.auth.identities import AuthUser
from typing import List
from aiohttp.client import ClientTimeout, ClientSession, _RequestContextManager
from requests.models import PreparedRequest
from urllib.parse import urlsplit, parse_qs

class OauthUser(AuthUser):
    token: str
    given_name: str
    family_name: str
    
    def __post_init__(self, data): 
        super(OauthUser, self).__post_init__(data)
        self.first_name = self.given_name
        self.last_name = self.family_name

class ExternalAuth(BaseAuthBackend):
    """ExternalAuth.

    is an abstract base to any External Auth backend, as Oauth2 or OpenId.
    """
    user_attribute: str = "user"
    username_attribute: str = "username"
    pwd_atrribute: str = "password"
    _service_name: str = "service"
    _user_mapping: Dict = {}

    def configure(self, app, router, handler):
        # add the callback url
        router = app.router
        self.base_url: str = ''
        self.authorize_uri: str = ''
        self.userinfo_uri: str = ''
        self._token_uri: str = ''
        # TODO: know the host we already are running
        self.login_failed_uri = AUTH_LOGIN_FAILED_URI
        self.redirect_uri = "{domain}/auth/{service}/callback/"
        # start login
        router.add_route(
            "*",
            "/api/v1/auth/{}/".format(self._service_name),
            self.authenticate,
            name="{}_api_login".format(self._service_name)
        )
        # finish login (callback)
        router.add_route(
            "GET",
            "/auth/{}/callback/".format(self._service_name),
            self.auth_callback,
            name="{}_complete_login".format(self._service_name)
        )
        # logout process
        router.add_route(
            "GET",
            "/api/v1/auth/{}/logout".format(self._service_name),
            self.logout,
            name="{}_api_logout".format(self._service_name)
        )
        router.add_route(
            "GET",
            "/auth/{}/logout".format(self._service_name),
            self.finish_logout,
            name="{}_complete_logout".format(self._service_name)
        )
        super(ExternalAuth, self).configure(app, router, handler)
        
    def get_domain(self, request: web.Request) -> str:
        absolute_uri = str(request.url)
        domain_url = absolute_uri.replace(str(request.rel_url), '')
        logging.debug(f'DOMAIN: {domain_url}')
        return domain_url
 
    def redirect(self, uri: str):
        """redirect.
            Making the redirection to External Auth Page.
        """
        logging.debug(f'{self.__class__.__name__} URI: {uri}')
        return web.HTTPFound(uri)

    def prepare_url(self, url: str, params: Dict = None):
        req = PreparedRequest()
        req.prepare_url(url, params)
        return req.url
        
    def home_redirect(self, request: web.Request, token: str = None, token_type: str = 'Bearer', **kwargs):
        logging.debug(f'Finish Auth URI::: {AUTH_REDIRECT_URI}')
        headers = {
            "x-authenticated": 'true'
        }
        req = PreparedRequest()
        params = {}
        if token:
            headers["x-auth-token"] = token
            headers["x-auth-token-type"] = token_type
            params = {
                "token" : token, "type":  token_type
            }
        url = self.prepare_url(AUTH_REDIRECT_URI, params)
        return web.HTTPFound(url, headers=headers)

    @abstractmethod
    async def authenticate(self, request: web.Request):
        """ Authenticate, refresh or return the user credentials."""
        pass
    
    @abstractmethod
    async def auth_callback(self, request: web.Request):
        """auth_callback, Finish method for authentication."""
        pass
    
    @abstractmethod
    async def logout(self, request: web.Request):
        """logout, forgot credentials and remove the user session."""
        pass
    
    @abstractmethod
    async def finish_logout(self, request: web.Request):
        """finish_logout, Finish Logout Method."""
        pass
    
    def build_user_info(self, userdata: Dict) -> Dict:
        # User ID:
        userid = userdata[self.userid_attribute]
        userdata['id'] = userid
        userdata[self.session_key_property] = userid
        # TODO: mapping
        for key, val in self._user_mapping.items():
            try:
                userdata[key] = userdata[val]
            except KeyError:
                pass
        return (userdata, userid)
    
    async def create_user(self, request: web.Request, user_id: Any, userdata: Any, token: str):
        # TODO: only creates after validation:
        data = None
        try:
            user = OauthUser(data=userdata)
            user.id = user_id
            user.auth_method = self._service_name
            user.access_token = token
            try:
                user.username = userdata[self.username_attribute]
            except KeyError:
                user.username = user_id
            user.token = token # issued token:
            # logging.debug(f'User Created > {user}')
            payload = {
                "user_id": user_id,
                **userdata
            }
            # saving Auth data.
            await self.remember(
                request, user_id, userdata, user
            )
            # Create the User session.
            jwt_token = self.create_jwt(data=payload)
            data = {
                "token": jwt_token,
                "access_token": token,
                **userdata
            }
        except Exception as err:
            logging.exception(err)
        finally:
            return data
    
    @abstractmethod
    async def check_credentials(self, request: web.Request):
        """Check the validity of the current issued credentials."""
        pass
    
    def base_url(self):
        return self.base_url

    def get(self, url, **kwargs) -> web.Response:
        """Perform an HTTP GET request."""
        return self.request(url, method=hdrs.METH_GET, **kwargs)

    def post(self, url, **kwargs) -> web.Response:
        """Perform an HTTP POST request."""
        return self.request(url, method=hdrs.METH_POST, **kwargs)
    
    async def request(self, url: str, method: str ='get', token: str = None, token_type: str = 'Bearer', **kwargs) -> web.Response:
        """
        request.
            connect to an http source using aiohttp
        """
        timeout = ClientTimeout(total=120)
        if 'headers' in kwargs:
            headers = kwargs['headers'].copy()
            del kwargs['headers']
        else:
            headers = {}
        if token:
            headers["Authorization"] = f"{token_type} {token}"
        if 'content-type' not in headers:
            headers['content-type'] = 'application/json'
            headers['Accept'] = 'application/json'
        response = None
        async with ClientSession(trust_env=True) as client:
            async with client.request(
                method,
                url,
                headers=headers,
                timeout=timeout,
                allow_redirects=True,
                **kwargs
            ) as response:
                if response.status == 200:
                    try:
                        return await response.json()
                    except aiohttp.client_exceptions.ContentTypeError:
                        resp = await response.read()
                        return parse_qs(resp.decode("utf-8"))
                else:
                    resp = await response.read()
                    raise Exception(f'Error getting Session Information: {resp}')


"""Django Session Backend.

Navigator Authentication using Django Session Backend
description: read the Django session from Redis Backend
and decrypt, after that, a session will be created.
"""
import base64
import rapidjson
import logging
import asyncio

# redis pool
import aioredis
from aiohttp import web
from .base import BaseAuthBackend
from navigator.exceptions import (
    NavException,
    UserDoesntExists,
    InvalidAuth
)
from navigator.conf import (
    SESSION_URL,
    SESSION_TIMEOUT,
    SECRET_KEY,
    SESSION_PREFIX,
    SESSION_KEY,
    AUTH_SESSION_OBJECT
)

# User Identity
from navigator.auth.identities import AuthUser, Column

class DjangoUser(AuthUser):
    """DjangoUser.
    
    user authenticated with Django Session (sessionid bearer).
    """
    sessionid: str = Column(required=True)



class DjangoAuth(BaseAuthBackend):
    """Django SessionID Authentication Handler."""

    def configure(self, app, router, handler):
        async def _setup_redis(app):
            self.redis = aioredis.from_url(
                    SESSION_URL,
                    decode_responses=True,
                    encoding='utf-8'
            )
            async def _close_redis(app):
                await self.redis.close()
            app.on_cleanup.append(_close_redis)
            return self.redis

        asyncio.get_event_loop().run_until_complete(_setup_redis(app))
        # executing parent configurations
        super(DjangoAuth, self).configure(app, router, handler)

    async def check_credentials(self, request):
        """ Authentication and create a session."""
        return True

    async def get_payload(self, request):
        id = None
        try:
            if "Authorization" in request.headers:
                try:
                    scheme, id = request.headers.get("Authorization").strip().split(" ")
                except ValueError:
                    raise web.HTTPForbidden(
                        reason="Invalid authorization Header",
                    )
                if scheme != self.scheme:
                    raise web.HTTPForbidden(
                        reason="Invalid Session scheme",
                    )
            elif "x-sessionid" in request.headers:
                id = request.headers.get("x-sessionid", None)
        except Exception as e:
            return None
        return id

    async def validate_session(self, key: str = None):
        try:
            async with await self.redis as redis:
                result = await redis.get("{}:{}".format(SESSION_PREFIX, key))
            if not result:
                raise Exception('Django Auth: non-existing Session')
            data = base64.b64decode(result)
            session_data = data.decode("utf-8").split(":", 1)
            user = rapidjson.loads(session_data[1])
            session = {
                "key": key,
                "session_id": session_data[0],
                self.user_property: user,
            }
            return session
        except Exception as err:
            logging.debug("Django Decoding Error: {}".format(err))
            raise

    async def validate_user(self, login: str = None):
        # get the user based on Model
        search = {self.userid_attribute: login}
        try:
            user = await self.get_user(**search)
            return user
        except UserDoesntExists as err:
            raise UserDoesntExists(f"User {login} doesn\'t exists")
        except Exception:
            raise
        return None

    async def authenticate(self, request):
        """ Authenticate against user credentials (django session id)."""
        try:
            sessionid = await self.get_payload(request)
            logging.debug(f"Session ID: {sessionid}")
        except Exception as err:
            raise NavException(err, state=400)
        if not sessionid:
            raise InvalidAuth(
                "Django Auth: Missing Credentials",
                state=401
            )
        else:
            try:
                data = await self.validate_session(
                    key=sessionid
                )
            except Exception as err:
                raise InvalidAuth(f"{err!s}", state=401)
            # making validation
            if not data:
                raise InvalidAuth("Django Auth: Missing User Info", state=403)
            try:
                u = data[self.user_property]
                username = u[self.userid_attribute]
            except KeyError as err:
                raise InvalidAuth(
                    f"Missing {self.userid_attribute} attribute: {err!s}",
                    state=401
                )
            try:
                user = await self.validate_user(
                    login=username
                )
            except UserDoesntExists as err:
                raise UserDoesntExists(err)
            except Exception as err:
                raise NavException(err, state=500)
            try:
                userdata = self.get_userdata(user)
                try:
                    # merging both session objects
                    userdata[AUTH_SESSION_OBJECT] = {
                        **userdata[AUTH_SESSION_OBJECT],
                        **data
                    }
                    usr = DjangoUser(data=userdata[AUTH_SESSION_OBJECT])
                    usr.id = sessionid
                    usr.sessionid = sessionid
                    usr.set(self.username_attribute, user[self.username_attribute])
                    logging.debug(f'User Created > {usr}')
                except Exception as err:
                    logging.exception(err)
                userdata[self.session_key_property] = sessionid
                # saving user-data into request:
                await self.remember(
                    request, sessionid, userdata, usr
                )
                payload = {
                    self.user_property: user[self.userid_attribute],
                    self.username_attribute: user[self.username_attribute],
                    self.userid_attribute: user[self.userid_attribute],
                    self.session_key_property: sessionid
                }
                token = self.create_jwt(
                    data=payload
                )
                return {
                    "token": token,
                    **userdata
                }
            except Exception as err:
                logging.exception(f'DjangoAuth: Authentication Error: {err}')
                return False



# -*- coding: utf-8 -*-
#!/usr/bin/env python3
import json
from aiohttp import web
from navigator.conf import (
    SESSION_KEY
)
from navigator.handlers import nav_exception_handler
from navigator.exceptions import (
    NavException,
    UserDoesntExists,
    InvalidAuth
)
from navigator.auth.sessions import get_session
from navigator.views import BaseView, BaseHandler
from asyncdb.utils.encoders import DefaultEncoder
from navigator.auth.models import User

class UserHandler(BaseView):

    async def session(self):
        session = None
        try:
            session = await get_session(self.request)
        except Exception as err:
            print(err)
            return self.critical(
                request=self.request,
                exception=err
            )
        return session

    async def get(self):
        """ Getting Session information."""
        session = await self.session()
        try:
            if not session:
                headers = {"x-status": "Empty", "x-message": "Invalid User Session"}
                return self.no_content(headers=headers)
            else:
                try:
                    id = session[SESSION_KEY]
                except KeyError:
                    return self.error('Invalid Session, missing Session ID')
                headers = {"x-status": "OK", "x-message": "Session OK"}
                userdata = dict(session)
                data = {
                    "session_id": id,
                    **userdata
                }
                if data:
                    return self.json_response(
                        response=data,
                        headers=headers
                    )
        except Exception as err:
            return self.error(
                self.request,
                exception=err
            )

    async def delete(self):
        """ Close and Delete User Session."""
        session = await self.session()
        try:
            app = self.request.app
            router = app.router
            session.invalidate()
            print(session)
        except Exception as err:
            print(err, err.__class__.__name__)
            return self.critical(
                request=self.request,
                exception=err,
                state=501
            )
        # return a redirect to LOGIN
        return web.HTTPFound(router["login"].url_for())

    async def put(self):
        """Re-login and re-authenticate..."""


class UserInfo(BaseHandler):

    async def session(self, request):
        session = None
        try:
            session = await get_session(request)
        except Exception as err:
            print(err)
            return self.critical(
                request=request,
                exception=err
            )
        return session

    async def profile(self, request):
        session = await self.session(request)
        print(session)
        if not session:
            headers = {"x-status": "Empty", "x-message": "Invalid User Session"}
            return self.no_content(headers=headers)
        else:
            try:
                sessionid = session['id']
            except KeyError:
                return self.error('Invalid Session, missing Session ID')
        # getting User information
        try:
            user_id = request["user_id"]
        except KeyError:
            info = session[sessionid]
            user_id = info['user_id']
        try:
            user = await User.get(user_id=user_id)
            return web.Response(
                text=user.json(ensure_ascii=True, indent=4),
                status=200,
                content_type="application/json"
            )
        except Exception as err:
            print(err)
            return self.critical(
                request=request,
                exception=err
            )

    async def logout(self, request):
        """ Close and Delete User Session."""
        session = await self.session(request)
        try:
            app = request.app
            router = app.router
            session.invalidate()
        except Exception as err:
            print(err, err.__class__.__name__)
            response = {
                "message": f"Exception on: {err.__class__.__name__}",
                "error": str(err)
            }
            args = {
                "status": 501,
                "content_type": "application/json",
                "text": json.dumps(response, cls=DefaultEncoder)
            }
            return web.Response(**args)
        # return a redirect to LOGIN
        # TODO: configure the return of LOGOUT
        return web.HTTPFound('/')



"""Connection Manager for Navigator."""

import logging
from multiprocessing.connection import wait
from aiohttp import web
from asyncdb import AsyncPool, AsyncDB
from typing import (
    Dict,
    List,
    Callable,
    Optional,
    Iterable
)
try:
    from settings.settings import TIMEZONE
except ImportError:
    # Timezone (For parsedate)
    TIMEZONE = "America/New_York"


class AbstractConnection(object):
    pool_based: bool = True
    timeout: int = 60
    _init: Callable = None
    _startup: Callable = None
    _shutdown: Callable = None

    def __init__(
        self,
        driver: str = "pg",
        dsn: str = "",
        name: str = "",
        init: Callable = None,
        startup: Callable = None,
        shutdown: Callable = None,
        **kwargs
    ):
        self.driver = driver
        if init:
            self._init = init
        if startup:
            self._startup = startup
        if shutdown:
            self._shutdown = shutdown

        if self.pool_based:
            self.conn = AsyncPool(self.driver, dsn=dsn, timeout=self.timeout, **kwargs)
        else:
            self.conn = AsyncDB(self.driver, dsn=dsn, timeout=self.timeout, **kwargs)
        if self._init:
            self.conn.setup_func = self._init

    def connection(self):
        return self.conn

    def configure(self, app: web.Application):
        app.on_startup.append(
            self.startup
        )
        app.on_shutdown.append(
            self.shutdown
        )
        app.on_cleanup.append(
            self.cleanup
        )
        

    def is_connected(self):
        return bool(self.conn.is_connected())
    
    async def cleanup(self, app: web.Application):
        pass

    async def startup(self, app: web.Application):
        if self.pool_based:
            await self.conn.connect()
        else:
            await self.conn.connection()
        if self._startup: # any callable will be launch on connection startup.
            await self._startup(self.conn)

    async def shutdown(self, app: web.Application):
        if self._shutdown:
            await self._shutdown(self.conn)
        logging.debug(" === Closing all connections === ")
        try:
            await self.conn.close()
        finally:
            logging.debug("Exiting ...")
        
        
class PostgresPool(AbstractConnection):
    driver: str = "pg"
    pool_based: bool = True
    timeout: int = 3600

    def __init__(
        self,
        dsn: str,
        name: str = "",
        init: Callable = None,
        startup: Callable = None,
        shutdown: Callable = None,
        **kwargs
    ):
        kwargs = {
            "min_size": 5,
            "server_settings": {
                "application_name": name,
                "client_min_messages": "notice",
                "max_parallel_workers": "48",
                "jit": "off",
                "statement_timeout": "60",
                "effective_cache_size": "2147483647"
            },
        }
        super(PostgresPool, self).__init__(self.driver, dsn, name, init, startup, shutdown, **kwargs)

    async def shutdown(self, app: web.Application):
        if self._shutdown:
            await self._shutdown(self.conn)
        logging.debug(" === Closing all connections === ")
        try:
            await self.conn.wait_close(gracefully=True, timeout=2)
        finally:
            logging.debug("Exiting ...")

class RedisPool(AbstractConnection):
    driver: str = "redis"
    pool_based: bool = True
    timeout: int = 60

    def __init__(
        self,
        dsn: str,
        name: str = "",
        init: Callable = None,
        startup: Callable = None,
        shutdown: Callable = None,
        **kwargs
    ):
        super(RedisPool, self).__init__(
            self.driver, dsn, name, init, startup, shutdown, **kwargs
        )
        
    async def shutdown(self, app: web.Application):
        pass

    async def cleanup(self, app: web.Application):
        if self._shutdown:
            await self._shutdown(self.conn)
        logging.debug(" === Closing REDIS === ")
        try:
            await self.conn.close()
        finally:
            logging.debug("Exiting ...")


"""GoogleAuth.

Description: Backend Authentication/Authorization using Google AUTH API.
"""
import logging
from aiohttp import web
from navigator.exceptions import (
    NavException,
    UserDoesntExists,
    InvalidAuth
)
from aiogoogle import Aiogoogle
from aiogoogle.auth.utils import create_secret
from navigator.conf import (
    GOOGLE_CLIENT_ID,
    GOOGLE_CLIENT_SECRET,
    GOOGLE_API_SCOPES
)
from .external import ExternalAuth


class GoogleAuth(ExternalAuth):
    """GoogleAuth.

    Authentication Backend using Google+aiogoogle.
    """
    user_attribute: str = "user"
    username_attribute: str = "username"
    pwd_atrribute: str = "password"
    _service_name: str = "google"

    def configure(self, app, router, handler):
        super(GoogleAuth, self).configure(app, router, handler)
        # TODO: build the callback URL and append to routes
        self._credentials = {
            "client_id": GOOGLE_CLIENT_ID,
            "client_secret": GOOGLE_CLIENT_SECRET,
            "scopes": GOOGLE_API_SCOPES,
            "redirect_uri": self.redirect_uri,
        }


    async def get_payload(self, request):
        pass

    async def authenticate(self, request):
        """ Authenticate, refresh or return the user credentials."""
        self._state = (
            create_secret()
        )  # Shouldn't be a global or a hardcoded variable. should be tied to a session or a user and shouldn't be used more than once
        self._nonce = (
            create_secret()
        )  # Shouldn't be a global or a hardcoded variable. should be tied to a session or a user and shouldn't be used more than once
        domain_url = self.get_domain(request)        
        self.redirect_uri = self.redirect_uri.format(domain=domain_url, service=self._service_name)
        self._credentials["redirect_uri"] = self.redirect_uri
        self.google = Aiogoogle(
            client_creds=self._credentials
        )
        if self.google.openid_connect.is_ready(self._credentials):
            uri = self.google.openid_connect.authorization_url(
                client_creds=self._credentials,
                state=self._state,
                nonce=self._nonce,
                access_type="offline",
                include_granted_scopes=True,
                # login_hint=user,
                prompt="select_account",
            )
            # Step A: redirect
            return self.redirect(uri)
        else:
            raise NavException(
                "Client doesn't have info for Authentication"
            )

    async def auth_callback(self, request: web.Request):
        try:
            if request.query.get('error'):
                error = {
                    "error": request.query.get("error"),
                    "error_description": request.query.get("error_description"),
                }
                logging.exception(f"Google Login Error: {err}, {err.state}")
                response = {
                    "message": "Google Login Error",
                    **error
                }
                return web.json_response(response, status=403)
        except Exception as err:
            logging.exception(f"Google Login Error: {err}, {err.state}")
            response = {
                "message": f"Google Login Error: {err}, {err.state}"
            }
            return web.json_response(response, status=501)
        if code:=request.query.get("code"):
            state = request.query.get("state")
            if state != self._state:
                response = {
                    "message": "Something wrong with Authentication State",
                    "error": "Authenticate Error"
                }
                return web.json_response(response, status=403)
            user_creds = await self.google.openid_connect.build_user_creds(
                grant=code,
                client_creds=self._credentials,
                nonce=self._nonce,
                verify=False,
            )
            # print(user_creds)
            userdata = await self.google.openid_connect.get_user_info(
                user_creds
            )
            try:
                print(userdata)
                id = userdata['id']
                access_token = user_creds['id_token_jwt']
                userdata[self.session_key_property] = id
                data = await self.create_user(request, id, userdata, access_token)
                return self.home_redirect(request, token=data['token'], token_type='Bearer')
            except Exception as err:
                logging.exception(f"Okta Auth Error: {err}")
                return self.redirect(uri=self.login_failed_uri)
        else:
            response = {
                "message": "Something wrong with Google Callback",
                "error": "Authenticate Error"
            }
            return web.json_response(response, status=403)

    async def logout(self, request):
        pass

    async def finish_logout(self, request):
        pass

    async def check_credentials(self, request):
        """ Authentication and create a session."""
        return True



"""
Abstract Class for Authorization Policies and decorators
"""
from aiohttp import web
from abc import ABC, abstractmethod
from functools import wraps

class AuthorizationPolicy(ABC):
    @abstractmethod
    async def permits(self, identity, permission, context=None):
        """Check user permissions.
        Return True if the identity is allowed the permission in the
        current context, else return False.
        """
        pass

    @abstractmethod
    async def is_authorized(self, identity):
        """Retrieve authorized user id.
        Return the user_id of the user identified by the identity
        or 'None' if no user exists related to the identity.
        """
        pass



"""Oauth.

Oauth is a Abstract Class with basic functionalities for all Oauth2 backends.
"""

import logging
import requests
from aiohttp import web
from .external import ExternalAuth
from navigator.exceptions import (
    NavException
)
from abc import abstractmethod
from typing import Dict


class OauthAuth(ExternalAuth):
    """OauthAuth.

    Description: Abstract Class for all Oauth2 backends.
    """
    _auth_code: str = 'code'
    
    @abstractmethod
    async def get_credentials(self, request: web.Request):
        pass
    
    async def authenticate(self, request: web.Request):
        """ Authenticate, refresh or return the user credentials."""
        try:
            domain_url = self.get_domain(request)        
            self.redirect_uri = self.redirect_uri.format(domain=domain_url, service=self._service_name)
            # Build the URL
            params = await self.get_credentials(request)
            url = self.prepare_url(self.authorize_uri, params)
            # Step A: redirect
            return self.redirect(url)
        except Exception as err:
            raise NavException(
                f"{self._service_name}: Client doesn't have info for Authentication: {err}"
            )
            
    def get_auth_response(self, request: web.Request) -> Dict:
        return dict(request.rel_url.query.items())

    def get_auth_code(self, response: Dict) -> str:
        try:
            code = response.get(self._auth_code)
        except KeyError:
            code = None
        if not code:
            raise RuntimeError(
                f"Auth Error: {self._service_name} Code not accessible"
            )
        return code



"""JWT Backend.

Navigator Authentication using JSON Web Tokens.
"""
import logging
import hashlib
import base64
import secrets
from .base import BaseAuthBackend
from navigator.exceptions import (
    NavException,
    FailedAuth,
    UserDoesntExists,
    InvalidAuth,
    ValidationError,
)
from navigator.conf import (
    AUTH_PWD_DIGEST,
    AUTH_PWD_ALGORITHM,
    AUTH_PWD_LENGTH,
    AUTH_PWD_SALT_LENGTH,
    AUTH_SESSION_OBJECT
)
# Authenticated Entity
from navigator.auth.identities import AuthUser

class BasicUser(AuthUser):
    """BasicAuth.
    
    Basic authenticated user.
    """
    pass


# "%s$%d$%s$%s" % (algorithm, iterations, salt, hash)


class BasicAuth(BaseAuthBackend):
    """Basic User/password Authentication."""

    user_attribute: str = "user"
    pwd_atrribute: str = "password"

    async def validate_user(self, login: str = None, password: str = None):
        # get the user based on Model
        try:
            search = {self.username_attribute: login}
            user = await self.get_user(**search)
        except UserDoesntExists as err:
            raise UserDoesntExists(
                f"User {login} doesn't exists"
            )
        except Exception:
            raise
        try:
            # later, check the password
            pwd = user[self.pwd_atrribute]
        except KeyError:
            raise ValidationError(
                'Missing Password attribute on User Account'
            )
        try:
            if self.check_password(pwd, password):
                # return the user Object
                return user
            else:
                raise FailedAuth(
                    "Basic Auth: Invalid Credentials"
                )
        except Exception:
            raise
        return None

    def set_password(
        self,
        password: str,
        token_num: int = 6,
        iterations: int = 80000,
        salt: str = None
    ):
        if not salt:
            salt = secrets.token_hex(token_num)
        key = hashlib.pbkdf2_hmac(
            AUTH_PWD_DIGEST,
            password.encode("utf-8"),
            salt.encode("utf-8"),
            iterations,
            dklen=AUTH_PWD_LENGTH,
        )
        hash = base64.b64encode(key).decode("utf-8").strip()
        return f"{AUTH_PWD_ALGORITHM}${iterations}${salt}${hash}"

    def check_password(self, current_password, password):
        try:
            algorithm, iterations, salt, hash = current_password.split("$", 3)
        except ValueError:
            raise InvalidAuth('Basic Auth: Invalid Password Algorithm')
        assert algorithm == AUTH_PWD_ALGORITHM
        compare_hash = self.set_password(
            password,
            iterations=int(iterations),
            salt=salt,
            token_num=AUTH_PWD_SALT_LENGTH
        )
        return secrets.compare_digest(current_password, compare_hash)

    async def get_payload(self, request):
        ctype = request.content_type
        if request.method == "GET":
            try:
                user = request.query.get(self.username_attribute, None)
                password = request.query.get(self.pwd_atrribute, None)
                return [user, password]
            except Exception:
                return [None, None]
        elif ctype in ("multipart/mixed", "multipart/form-data", "application/x-www-form-urlencoded"):
            data = await request.post()
            if len(data) > 0:
                user = data.get(self.username_attribute, None)
                password = data.get(self.pwd_atrribute, None)
                return [user, password]
            else:
                return [None, None]
        elif ctype == "application/json":
            try:
                data = await request.json()
                user = data[self.username_attribute]
                password = data[self.pwd_atrribute]
                return [user, password]
            except Exception:
                return [None, None]
        else:
            return [None, None]

    async def authenticate(self, request):
        """ Authenticate, refresh or return the user credentials."""
        try:
            user, pwd = await self.get_payload(request)
        except Exception as err:
            raise NavException(err, state=400)
        if not pwd and not user:
            raise InvalidAuth(
                "Basic Auth: Invalid Credentials",
                state=401
            )
        else:
            # making validation
            try:
                user = await self.validate_user(login=user, password=pwd)
            except FailedAuth as err:
                raise FailedAuth(err)
            except UserDoesntExists as err:
                raise UserDoesntExists(err)
            except (ValidationError, InvalidAuth) as err:
                raise InvalidAuth(err, state=401)
            except Exception as err:
                raise NavException(
                    err, state=500
                )
            try:
                userdata = self.get_userdata(user)
                username = user[self.username_attribute]
                id = user[self.userid_attribute]
                userdata[self.username_attribute] = username
                userdata[self.session_key_property] = username
                usr = BasicUser(data=userdata[AUTH_SESSION_OBJECT])
                usr.id = id
                usr.set(self.username_attribute, username)
                logging.debug(f'User Created > {usr}')
                payload = {
                    self.user_property: user[self.userid_attribute],
                    self.username_attribute: username,
                    "user_id": id,
                    self.session_key_property: username
                }
                # Create the User session and returned.
                token = self.create_jwt(data=payload)
                usr.access_token = token
                await self.remember(
                    request, username, userdata, usr
                )
                return {
                    "token": token,
                    **userdata
                }
            except Exception as err:
                logging.exception(f'DjangoAuth: Authentication Error: {err}')
                return False

    async def check_credentials(self, request):
        """ Using for check the user credentials to the backend."""
        pass



"""OktaAuth.

Description: Backend Authentication/Authorization using Okta Service.
"""
import logging
from aiohttp import web
from .oauth import OauthAuth
import requests
from navigator.exceptions import (
    NavException
)
from navigator.conf import (
    OKTA_CLIENT_ID,
    OKTA_CLIENT_SECRET,
    OKTA_DOMAIN,
    OKTA_APP_NAME,
    AUTH_REDIRECT_URI
)
from okta_jwt_verifier import JWTVerifier


async def is_token_valid(token, issuer, client_id):
    jwt_verifier = JWTVerifier(issuer, client_id, 'api://default')
    try:
        await jwt_verifier.verify_access_token(token)
        return True
    except Exception:
        return False


async def is_id_token_valid(token, issuer, client_id, nonce):
    jwt_verifier = JWTVerifier(issuer, client_id, 'api://default')
    try:
        await jwt_verifier.verify_id_token(token, nonce=nonce)
        return True
    except Exception:
        return False


class OktaAuth(OauthAuth):
    """OktaAuth.

    Description: Authentication Backend using Third-party Okta Service.
    """
    user_attribute: str = "user"
    username_attribute: str = "email"
    userid_attribute: str = "sub"
    pwd_atrribute: str = "password"
    _service_name: str = "okta"

    def configure(self, app, router, handler):
        super(OktaAuth, self).configure(app, router, handler) # first, configure parents
        
        # auth paths.
        self.base_url = f"https://{OKTA_DOMAIN}/"
        self.authorize_uri = f"https://{OKTA_DOMAIN}/oauth2/default/v1/authorize"
        self.userinfo_uri = f"https://{OKTA_DOMAIN}/oauth2/default/v1/userinfo"
        self._issuer = f"https://{OKTA_DOMAIN}/oauth2/default"
        self._token_uri = f"https://{OKTA_DOMAIN}/oauth2/default/v1/token"
        self._introspection_uri = f"https://{OKTA_DOMAIN}/oauth2/default/v1/introspect"

    async def get_credentials(self, request: web.Request):
        APP_STATE = 'ApplicationState'
        self.nonce = 'SampleNonce'
        domain_url = self.get_domain(request)        
        self.redirect_uri = self.redirect_uri.format(domain=domain_url, service=self._service_name)
        qs = {
            "client_id": f"{OKTA_CLIENT_ID}",
            # "client_secret": f"{OKTA_CLIENT_SECRET}",
            "redirect_uri": self.redirect_uri,
            'scope': "openid email profile",
            'state': APP_STATE,
            'nonce': self.nonce,
            'response_type': 'code',
            'response_mode': 'query',
        }
        return qs

    async def auth_callback(self, request: web.Request):
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        domain_url = self.get_domain(request)        
        self.redirect_uri = self.redirect_uri.format(domain=domain_url, service=self._service_name)
        code = request.query.get("code")
        if not code:
            response = {
                "message": "Auth Error: Okta Code not accessible"
            }
            return web.json_response(response, status=403)
        # B.- processing the code
        query_params = {
            'grant_type': 'authorization_code',
            'code': code,
            'redirect_uri': self.redirect_uri
        }
        query_params = requests.compat.urlencode(query_params)
        try:
            exchange = requests.post(
                self._token_uri,
                headers=headers,
                data=query_params,
                auth=(
                    OKTA_CLIENT_ID,
                    OKTA_CLIENT_SECRET
                ),
            ).json()
        except Exception as err:
            response = {
                "message": f"Okta: Error Getting User Profile information: {err}"
            }
            return web.json_response(response, status=403)
        # Get tokens and validate
        if not exchange.get("token_type"):
            response = {
                "message": "Okta: Unsupported token type. Should be 'Bearer'."
            }
            return web.json_response(response, status=403)
        # user data
        access_token = exchange["access_token"]
        id_token = exchange["id_token"]

        if not await is_token_valid(access_token, self._issuer, OKTA_CLIENT_ID):
            response = {
                "message": "Okta: Access Token Invalid."
            }
            return web.json_response(response, status=403)

        if not await is_id_token_valid(id_token, self._issuer, OKTA_CLIENT_ID, self.nonce):
            response = {
                "message": "Okta: ID Token Invalid."
            }
            return web.json_response(response, status=403)

        # Authorization flow successful, get userinfo and login user
        try:
            data = requests.get(
                self.userinfo_uri,
                headers={'Authorization': f'Bearer {access_token}'}
            ).json()
            userdata, uid = self.build_user_info(data)
            # get user data
            # TODO: Optional: get User info from Nav
            data = await self.create_user(request, uid, userdata, access_token)
            return self.home_redirect(request, token=data['token'], token_type='Bearer')
        except Exception as err:
            logging.exception(f"Okta Auth Error: {err}")
            return self.redirect(uri=self.login_failed_uri)

    async def logout(self, request):
        pass

    async def finish_logout(self, request):
        pass

    async def check_credentials(self, request):
        """ Authentication and create a session."""
        return True



import slugify
from asyncdb.models import Model, Column, Field
from dataclasses import dataclass, InitVar
from typing import (
    Any,
    List,
    Optional,
    Dict
)


class Group(Model):
    """Group.
    
    Association (group) were users belongs to.
    """
    group: str = Column(required=True)

    class Meta:
        strict = True
        frozen = False
        connection = None

# create a Guest Group.
Guest = Group(group = 'guest')


class Organization(Model):
    org_id: str
    organization: str
    slug: str
    
    def __post_init__(self) -> None:
        super(Organization, self).__post_init__()
        if not self.slug:
            self.slug = slugify(self.organization)

    class Meta:
        strict = True
        frozen = False
        connection = None

class Program(Model):
    program_id: int
    program_name: str
    program_slug: str
    
    def __post_init__(self) -> None:
        super(Program, self).__post_init__()
        if not self.program_slug:
            self.program_slug = slugify(self.program_name)

    class Meta:
        strict = True
        frozen = False
        connection = None

class Identity(Model):
    """Identity.
    
    Describe an Authenticated Entity on Navigator.
    """
    id: Any = Column(required=True)
    auth_method: str = None
    access_token: Optional[str] = None
    enabled: bool = Column(required=True, default=True)
    data: InitVar[Dict] = Column(required=False, default_factory=dict)
    is_authenticated: bool = Column(equired=False, default=False)
    userdata: Dict  = Column(required=False, default_factory={})

    def __post_init__(self, data):
        self.userdata = data
        for key, value in data.items():
            self.create_field(key, value)
            
    def create_field(self, name: str, value: Any) -> None:
        # create a new Field on Model (when strict is False).            
        f = Field(required=False, default=value)
        f.name = name
        f.type = type(value)
        self.__columns__[name] = f
        setattr(self, name, value)

    def set(self, name: str, value: Any) -> None:
        # alias for "create_field"
        self.create_field(name, value)

    class Meta:
        strict = False
        frozen = False
        connection = None


class AuthUser(Identity):
    """AuthUser

    Model for any Authenticated User.
    """
    first_name: str
    last_name: str
    name: str
    email: str
    username: str
    groups: List[Group] = Column(required=False, default_factory=list)
    organizations: List[Organization] = Column(required=False, default_factory=list)
    superuser: bool = Column(required=True, default=False)
    
    def __post_init__(self, data) -> None:
        super(AuthUser, self).__post_init__(data)
        if self.groups is not None:
            groups = self.groups.copy()
            self.groups = []
            for group in groups:
                self.add_group(group)
        else:
            self.groups = []
        self.organizations = []

    """
    User Methods.
    """
    def add_group(self, group: Group):
        if isinstance(group, str):
            group = Group(group=group)
        self.groups.append(group)




import jwt
from aiohttp import web
from typing import (
    Optional,
    Coroutine,
    Tuple
)
from navconfig import config
from navigator.conf import (
    SESSION_TIMEOUT,
    SECRET_KEY,
    CREDENTIALS_REQUIRED,
    SESSION_USER_PROPERTY,
    SECRET_KEY,
    JWT_ALGORITHM
)
from .abstract import base_middleware


class jwt_middleware(base_middleware):
    def __init__(
        self,
        user_fn: Coroutine,
        user_property: str = SESSION_USER_PROPERTY,
        exclude_routes: Optional[Tuple] = tuple(),
        jwt_algorithm: str = 'HS256'
    ):
        """
        Simple Middleware to decrypt JWT tokens and return the payload to a
        Callable.

        Args:
            user_fn: any User Callable Coroutine, to process the received token.
              TODO: if not, can we use own token factory.
            user_property: the User Object returned by callback is saved
              on this property under the request.
            exclude_routes: any path that needs to be excluded for Token Auth
        Returns:
            handler if User exists, HTTP Forbidden if callback returns false.
            HTTP Unauthorized if Token is missing (only if credential required is TRUE)
        """
        if not callable(user_fn):
            raise RuntimeError(
                f"User Function {user_fn!s} need to be an Asyncio Coroutine."
            )
            # TODO: adds a generic list of exclude routes
        self._fn = user_fn
        self.jwt_algorithm = jwt_algorithm

        if exclude_routes is None:
            self.exclude_routes = config.get('EXCLUDED_ROUTES', tuple())
        else:
            self.exclude_routes = exclude_routes
        # user property
        self.user_property = user_property

    async def middleware(self, app, handler):
        @web.middleware
        async def middleware(request):
            if await self.valid_routes(request):
                return await handler(request)
            try:
                token, scheme = self.get_authorization_header(request, scheme = 'Bearer')
            except KeyError as err:
                token = None
            if CREDENTIALS_REQUIRED is True:
                if not token:
                    raise web.HTTPForbidden(
                        reason='Token Auth: Invalid authorization Token',
                    )
                try:
                    # process token:
                    payload = jwt.decode(
                        token,
                        SECRET_KEY,
                        algorithms=[self.jwt_algorithm]
                    )
                    if not payload:
                        raise web.HTTPForbidden(
                            reason="Invalid authorization Token"
                        )
                except (jwt.DecodeError) as err:
                    raise web.HTTPBadRequest(
                        reason=f"JWT: Invalid Token: {err}"
                    )
                except jwt.ExpiredSignatureError as err:
                    raise web.HTTPBadRequest(
                        reason=f"JWT: Expired Token or bad signature: {err}"
                    )
                except Exception as err:
                    raise web.HTTPBadRequest(
                        reason=f"JWT Token Decryption Error: {err}"
                    )
                try:
                    user = await self._fn(payload, request)
                    if user:
                        request[self.user_property] = user
                        request.user = user
                        print(user)
                    else:
                        raise web.HTTPForbidden(
                            reason='Access Restricted'
                        )
                except Exception as err:
                    raise web.HTTPBadRequest(
                        reason=f'Token Auth: Exception on Callable Return {err!s}'
                    )
            return await handler(request)
        return middleware
# async def jwt_middleware(app, handler):
#     async def middleware(request):
#         request.user = None
#         jwt_token = request.headers.get("Authorization", None)
#         if jwt_token:
#             try:
#                 payload = jwt.decode(
#                     jwt_token,
#                     SECRET_KEY,
#                     algorithms=[JWT_ALGORITHM]
#                 )
#                 print(payload)
#                 request.user = payload
#             except (jwt.DecodeError, jwt.ExpiredSignatureError) as err:
#                 print(err)
#                 return web.json_response({"message": "Invalid Token"}, status=400)
#         return await handler(request)

#     return middleware



import logging
import jwt
import asyncio
from typing import (
    List,
    Dict,
    Optional,
    Callable
)
from abc import ABC, abstractmethod
from aiohttp import web, hdrs
from datetime import datetime, timedelta
from asyncdb.models import Model
from cryptography import fernet
from functools import partial, wraps
from concurrent.futures import ThreadPoolExecutor
import base64
from navigator.conf import (
    AUTH_USER_MODEL,
    AUTH_GROUP_MODEL,
    AUTH_SESSION_OBJECT,
    AUTH_USERNAME_ATTRIBUTE,
    JWT_ALGORITHM,
    USER_MAPPING,
    SESSION_TIMEOUT,
    CREDENTIALS_REQUIRED,
    SESSION_KEY,
    SECRET_KEY,
    SESSION_USER_PROPERTY
)
from navigator.exceptions import (
    NavException,
    UserDoesntExists,
    InvalidAuth,
    FailedAuth,
    AuthExpired
)
from aiohttp.web_urldispatcher import SystemRoute
from navigator.auth.sessions import get_session, new_session

# Authenticated Identity
from navigator.auth.identities import Identity

exclude_list = (
    "/static/",
    "/api/v1/login",
    "/api/v1/logout",
    "/login",
    "/logout",
    "/signin",
    "/signout",
    "/_debug/",
)


class BaseAuthBackend(ABC):
    """Abstract Base for Authentication."""
    user_attribute: str = "user"
    password_attribute: str = "password"
    userid_attribute: str = "user_id"
    username_attribute: str = AUTH_USERNAME_ATTRIBUTE
    session_key_property: str = SESSION_KEY
    credentials_required: bool = CREDENTIALS_REQUIRED
    scheme: str = "Bearer"
    session_timeout: int = int(SESSION_TIMEOUT)
    _service: str = None

    def __init__(
        self,
        user_attribute: str = None,
        userid_attribute: str = None,
        password_attribute: str = None,
        credentials_required: bool = False,
        authorization_backends: tuple = (),
        **kwargs,
    ):
        self._service = self.__class__.__name__
        self._session = None
        # force using of credentials
        self.credentials_required = credentials_required
        self._credentials = None
        self.user_property = SESSION_USER_PROPERTY
        if user_attribute:
            self.user_attribute = user_attribute
        if password_attribute:
            self.password_attribute = password_attribute
        if userid_attribute:
            self.userid_attribute = userid_attribute
        self.username_attribute = AUTH_USERNAME_ATTRIBUTE
        # authentication scheme
        try:
            self.scheme = kwargs["scheme"]
        except KeyError:
            pass
        # configuration Authorization Backends:
        self._authz_backends: List = authorization_backends
        # user and group models
        # getting User and Group Models
        self.user_model: Model = kwargs["user_model"]
        # user mapping
        self.user_mapping = USER_MAPPING
        if not SECRET_KEY:
            fernet_key = fernet.Fernet.generate_key()
            self.secret_key = base64.urlsafe_b64decode(fernet_key)
        else:
            self.secret_key = SECRET_KEY
        # starts the Executor
        self.executor = ThreadPoolExecutor(max_workers=1)

    async def on_startup(self, app: web.Application):
        pass
    
    async def on_cleanup(self, app: web.Application):
        pass

    async def get_user(self, **search):
        """Getting User Object."""
        # TODO: getting Groups based on User
        try:
            user = await self.user_model.get(**search)
        except Exception as err:
            logging.error(f"Error getting User {search!s}")
            raise Exception(err)
        # if not exists, return error of missing
        if not user:
            raise UserDoesntExists(f"User doesnt exists")
        return user

    def get_userdata(self, user):
        userdata = {}
        for name, item in self.user_mapping.items():
            if name != self.password_attribute:
                userdata[name] = user[item]
        if AUTH_SESSION_OBJECT:
            return {
                AUTH_SESSION_OBJECT: userdata
            }
        return userdata

    def configure(self, app, router, handler):
        """Base configuration for Auth Backends, need to be extended
        to create Session Object."""
        pass

    async def remember(
            self,
            request: web.Request,
            identity: str,
            userdata: Dict,
            user: Identity
        ):
        """
        Saves User Identity into request Object.
        """
        try:
            request[self.session_key_property] = identity
            # request[self.user_property] = userdata
            # saving the user
            request.user = user
            # Session:
            try:
                session = await new_session(request, userdata)
                user.is_authenticated = True # if session, then, user is authenticated.
                # which Auth Method:
                # user.auth_method = self._service
                session[self.session_key_property] = identity
                session['user'] = session.encode(user)
                request['session'] = session
            except Exception as err:
                raise web.HTTPForbidden(
                    reason=f"Error Creating User Session: {err!s}"
                )
            # to allowing request.user.is_authenticated
        except Exception as err:
            logging.exception(err)

    async def authorization_backends(self, app, handler, request):
        try:
            if isinstance(request.match_info.route, SystemRoute):  # eg. 404
                return True
        except Exception as err:
            logging.error(err)
        # avoid authorization on exclude list
        if request.path in exclude_list:
            return True
        # avoid authorization backend on excluded methods:
        if request.method == hdrs.METH_OPTIONS:
            return True
        try:
            # logic for authorization backends
            for backend in self._authz_backends:
                if backend.check_authorization(request):
                    return True
        except Exception as err:
            logging.error(err)
        return None

    def create_jwt(
        self,
        issuer: str = None,
        expiration: int = None,
        data: dict = None
    ) -> str:
        """Creation of JWT tokens based on basic parameters.
        issuer: for default, urn:Navigator
        expiration: in seconds
        **kwargs: data to put in payload
        """
        if not expiration:
            expiration = self.session_timeout
        if not issuer:
            issuer = "urn:Navigator"
        payload = {
            "exp": datetime.utcnow() + timedelta(seconds=expiration),
            "iat": datetime.utcnow(),
            "iss": issuer,
            **data,
        }
        jwt_token = jwt.encode(
            payload,
            self.secret_key,
            JWT_ALGORITHM,
        )
        return jwt_token

    def decode_token(self, request, issuer: str = None):
        jwt_token = None
        tenant = None
        id = None
        payload = None
        if not issuer:
            issuer = "urn:Navigator"
        if "Authorization" in request.headers:
            try:
                scheme, id = (
                    request.headers.get(hdrs.AUTHORIZATION).strip().split(" ", 1)
                )
            except ValueError:
                raise NavException(
                    "Invalid authorization Header",
                    state=400
                )
            if scheme != self.scheme:
                raise NavException(
                    "Invalid Authorization Scheme",
                    state=400
                )
            try:
                tenant, jwt_token = id.split(":")
            except Exception:
                # normal Token:
                jwt_token = id
            logging.debug(f'Session Token: {jwt_token}')
            try:
                payload = jwt.decode(
                    jwt_token,
                    self.secret_key,
                    algorithms=[JWT_ALGORITHM],
                    iss=issuer,
                    leeway=30,
                )
                logging.info(f"Decoded Token: {payload!s}")
                return [tenant, payload]
            except jwt.exceptions.ExpiredSignatureError as err:
                raise AuthExpired(f"Credentials Expired: {err!s}")
            except jwt.exceptions.InvalidSignatureError as err:
                raise AuthExpired(f"Signature Failed or Expired: {err!s}")
            except jwt.exceptions.DecodeError as err:
                raise FailedAuth(f"Token Decoding Error: {err}")
            except jwt.exceptions.InvalidTokenError as err:
                raise InvalidAuth(f"Invalid authorization token {err!s}")
            except Exception as err:
                raise NavException(err, state=501)
        else:
            return [tenant, payload]

    @abstractmethod
    async def check_credentials(self, request):
        """ Authenticate against user credentials (token, user/password)."""
        pass

    def threaded_function(self, func: Callable, loop: asyncio.AbstractEventLoop = None, threaded: bool = True):
        """Wraps a Function into an Executor Thread."""
        @wraps(func)
        async def _wrap(*args, loop: asyncio.AbstractEventLoop = None, **kwargs):
            result = None
            if loop is None:
                loop = asyncio.new_event_loop()
            try:
                if threaded:
                    fn = partial(func, *args, **kwargs)
                    result = await loop.run_in_executor(
                        self.executor, fn
                    )
                else:
                    result = await func(*args, **kwargs)
                return result
            except Exception as err:
                logging.exception(err)
        return _wrap


"""ADFSAuth.

Description: Backend Authentication/Authorization using Okta Service.
"""
import rapidjson
import logging
import asyncio
import base64
from aiohttp import web, hdrs
from .external import ExternalAuth
from typing import List, Dict, Any
import jwt
# needed by ADFS
from xml.etree import ElementTree
import requests
import requests.adapters
from urllib3.util.retry import Retry
from cryptography.hazmat.backends.openssl.backend import backend
from cryptography.x509 import load_der_x509_certificate

from navigator.exceptions import (
    NavException,
    UserDoesntExists,
    InvalidAuth,
    FailedAuth
)
from navigator.conf import (
    ADFS_SERVER,
    ADFS_CLIENT_ID,
    ADFS_RELYING_PARTY_ID,
    ADFS_TENANT_ID,
    ADFS_RESOURCE,
    ADFS_AUDIENCE,
    ADFS_ISSUER,
    USERNAME_CLAIM,
    GROUP_CLAIM,
    ADFS_CLAIM_MAPPING,
    ADFS_CALLBACK_REDIRECT_URL,
    AZURE_AD_SERVER
)


def load_certs(response):
    """Extract token signing certificates."""
    xml_tree = ElementTree.fromstring(response.content)
    cert_nodes = xml_tree.findall(
            "./{urn:oasis:names:tc:SAML:2.0:metadata}RoleDescriptor"
            "[@{http://www.w3.org/2001/XMLSchema-instance}type='fed:SecurityTokenServiceType']"
            "/{urn:oasis:names:tc:SAML:2.0:metadata}KeyDescriptor[@use='signing']"
            "/{http://www.w3.org/2000/09/xmldsig#}KeyInfo"
            "/{http://www.w3.org/2000/09/xmldsig#}X509Data"
            "/{http://www.w3.org/2000/09/xmldsig#}X509Certificate"
    )
    signing_certificates = [node.text for node in cert_nodes]
    new_keys = []
    for cert in signing_certificates:
        logging.debug("Loading public key from certificate: %s", cert)
        cert_obj = load_der_x509_certificate(
            base64.b64decode(cert), backend
        )
        new_keys.append(
            cert_obj.public_key()
        )
    return new_keys

class ADFSAuth(ExternalAuth):
    """ADFSAuth.

    Description: Authentication Backend using
    Active Directory Federation Service (ADFS).
    """
    _service_name: str = "adfs"
    user_attribute: str = "user"
    username_attribute: str = "username"
    pwd_atrribute: str = "password"
    version = 'v1.0'
    _user_mapping: Dict = {
        'email': 'userPrincipalName',
        'given_name': 'givenName',
        'family_name': 'surname',
        'name': 'displayName'
    }

    def configure(self, app, router, handler):
        super(ADFSAuth, self).configure(app, router, handler)
        # URIs:
        if ADFS_TENANT_ID:
            self.server = AZURE_AD_SERVER
            self.tenant_id = ADFS_TENANT_ID
            self.username_claim = "upn"
            self.groups_claim = "groups"
            self.claim_mapping = ADFS_CLAIM_MAPPING
        else:
            self.server = ADFS_SERVER
            self.tenant_id = 'adfs'
            self.username_claim = USERNAME_CLAIM
            self.groups_claim = GROUP_CLAIM
            self.claim_mapping = ADFS_CLAIM_MAPPING
        
        self.base_uri = f"https:://{self.server}/"
        self.end_session_endpoint = f"https://{self.server}/{self.tenant_id}/ls/?wa=wsignout1.0"
        self._issuer = f"https://{self.server}/{self.tenant_id}/services/trust"
        self.authorize_uri = f"https://{self.server}/{self.tenant_id}/oauth2/authorize/"
        self._token_uri = f"https://{self.server}/{self.tenant_id}/oauth2/token"

    async def authenticate(self, request: web.Request):
        """ Authenticate, refresh or return the user credentials.

        Description: This function returns the ADFS authorization URL.
        """
        domain_url = self.get_domain(request)        
        if ADFS_CALLBACK_REDIRECT_URL:
            self.redirect_uri = ADFS_CALLBACK_REDIRECT_URL
        else:
            self.redirect_uri = self.redirect_uri.format(domain=domain_url, service=self._service_name)
        print('REDIRECT: ', self.redirect_uri)
        try:
            self.state = base64.urlsafe_b64encode(self.redirect_uri.encode()).decode()
            query_params = {
                "client_id": ADFS_CLIENT_ID,
                "response_type": "code",
                "redirect_uri": self.redirect_uri,
                "resource": ADFS_RESOURCE,
                "response_mode": "query",
                "state": self.state,
                "scope": "openid,offline_access"
            }
            login_url = "{base_uri}?{query_params}".format(
                base_uri=self.authorize_uri,
                query_params=requests.compat.urlencode(query_params)
            )
            # Step A: redirect
            return self.redirect(login_url)
        except Exception as err:
            print('HERE: ', err)
            raise NavException(
                f"Client doesn't have info for ADFS Authentication: {err}"
            )

    async def auth_callback(self, request: web.Request):
        absolute_uri = str(request.url)
        DOMAIN_URL = absolute_uri.replace(str(request.rel_url), '')          
        if ADFS_LOGIN_REDIRECT_URL:
            self.redirect_uri = ADFS_LOGIN_REDIRECT_URL
        else:
            self.redirect_uri = f"{DOMAIN_URL}/auth/adfs/callback/"
        try:
            auth_response = dict(request.rel_url.query.items())
            authorization_code = auth_response['code']
            state = auth_response['state']
            request_id = auth_response['client-request-id']
        except Exception as err:
            print(err)
            raise NavException(
                f"ADFS: Invalid Callback response: {err}"
            )
        print(authorization_code, state, request_id)
        logging.debug("Received authorization token: " + authorization_code)
        # getting an Access Token
        query_params = {
            "code": authorization_code,
            "client_id": ADFS_CLIENT_ID,
            "grant_type": "authorization_code",
            "redirect_uri": self.redirect_uri,
            "scope": "https://graph.microsoft.com/.default"
        }
        headers = {
            "Content-Type": "application/x-www-form-urlencoded"
        }
        try:
            exchange = await self.post(self._token_uri, data=query_params, headers=headers)
            if 'error' in exchange:
                error = exchange.get('error')
                desc = exchange.get('error_description')
                message = f"Azure {error}: {desc}¡"
                logging.exception(message)
                raise web.HTTPForbidden(
                    reason=message
                )
            else:
                ## processing the exchange response:
                access_token = exchange["access_token"]
                token_type = exchange["token_type"] # ex: Bearer
                id_token = exchange["id_token"]
                logging.debug(f"Received access token: {access_token}")
                # getting user information:
                try:
                    data = await self.get(url=self.userinfo_uri, token=access_token, token_type=token_type)
                    # build user information:
                    userdata, uid = self.build_user_info(data)
                    userdata['id_token'] = id_token
                    data = await self.create_user(request, uid, userdata, access_token)
                except Exception as err:
                    logging.exception(f'ADFS: Error getting User information: {err}')
                    raise web.HTTPForbidden(
                        reason=f"ADFS: Error with User Information: {err}"
                    )
        except Exception as err:
            raise web.HTTPForbidden(
                reason=f"ADFS: Invalid Response from Server {err}."
            )

    async def logout(self, request):
        # first: removing the existing session
        # second: redirect to SSO logout
        logging.debug(f'ADFS LOGOUT URI: {self.end_session_endpoint}')
        return web.HTTPFound(self.end_session_endpoint)

    async def finish_logout(self, request):
        pass

    async def check_credentials(self, request):
        """ Authentication and create a session."""
        return True



"""AzureAuth.

Description: Backend Authentication/Authorization using Microsoft authentication.
"""
import logging
from aiohttp import web
from navigator.exceptions import (
    NavException,
    UserDoesntExists,
    InvalidAuth
)
from navigator.conf import (
    AZURE_ADFS_CLIENT_ID,
    AZURE_ADFS_CLIENT_SECRET,
    AZURE_ADFS_SCOPES,
    AZURE_ADFS_DOMAIN,
    AZURE_ADFS_TENANT_ID
)
import msal
import json
import jwt
from .external import ExternalAuth
from typing import Any, Dict
from msal.authority import (
    AuthorityBuilder,
    AZURE_PUBLIC
)

logging.getLogger("msal").setLevel(logging.INFO)

class AzureAuth(ExternalAuth):
    """AzureAuth.

    Authentication Backend for Microsoft Online Services.
    """
    user_attribute: str = "user"
    username_attribute: str = "username"
    userid_attribute: str = 'userPrincipalName'
    pwd_atrribute: str = "password"
    _service_name: str = "azure"
    _user_mapping: Dict = {
        'email': 'userPrincipalName',
        'given_name': 'givenName',
        'family_name': 'surname',
        'name': 'displayName'
    }

    def configure(self, app, router, handler):
        super(AzureAuth, self).configure(app, router, handler)

        # TODO: build the callback URL and append to routes
        self.base_url: str = 'https://login.microsoftonline.com/{tenant}'.format(
            tenant=AZURE_ADFS_TENANT_ID
        )
        self.authorize_uri = "https://login.microsoftonline.com/{tenant}/oauth2/v2.0/authorize".format(
            tenant=AZURE_ADFS_TENANT_ID
        )
        self.userinfo_uri = "https://graph.microsoft.com/v1.0/me"
        
        # issuer:
        self._issuer = "https://login.microsoftonline.com/{tenant}".format(
            tenant=AZURE_ADFS_TENANT_ID
        )
        self._token_uri = "https://login.microsoftonline.com/{tenant}/oauth2/v2.0/token".format(
            tenant=AZURE_ADFS_TENANT_ID
        )
        self.authority = AuthorityBuilder(AZURE_PUBLIC, "contoso.onmicrosoft.com")
        self.users_info = "https://graph.microsoft.com/v1.0/users"


    async def get_payload(self, request):
        ctype = request.content_type
        if request.method == "POST":
            if ctype in ("multipart/mixed", "multipart/form-data", "application/x-www-form-urlencoded"):
                data = await request.post()
                if len(data) > 0:
                    user = data.get(self.username_attribute, None)
                    password = data.get(self.pwd_atrribute, None)
                    return [user, password]
                else:
                    return [None, None]
            elif ctype == "application/json":
                try:
                    data = await request.json()
                    user = data[self.username_attribute]
                    password = data[self.pwd_atrribute]
                    return [user, password]
                except Exception:
                    return [None, None]
        else:
            return [None, None]

    
    async def get_cache(self, request: web.Request, state: str):
        cache = msal.SerializableTokenCache()
        result = None
        async with await request.app['redis'].acquire() as redis:
            result = await redis.get(f'azure_cache_{state}')
        if result:
            cache.deserialize(result)
        return cache
    
    async def save_cache(self, request: web.Request, state: str, cache: msal.SerializableTokenCache):
        if cache.has_state_changed:
            result = cache.serialize()
            async with await request.app['redis'].acquire() as redis:
                await redis.setex(f'azure_cache_{state}', result, timeout=120)
            
    def get_msal_app(self):
        authority = self._issuer if self._issuer else self.authority
        return msal.ConfidentialClientApplication(
            AZURE_ADFS_CLIENT_ID,
            authority=authority,
            client_credential = AZURE_ADFS_CLIENT_SECRET,
            validate_authority=True
            # token_cache=cache
        )
        
    def get_msal_client(self):
        authority = self._issuer if self._issuer else self.authority
        return msal.ClientApplication(
            AZURE_ADFS_CLIENT_ID,
            authority=authority,
            client_credential = AZURE_ADFS_CLIENT_SECRET,
            validate_authority=True
        )

    async def authenticate(self, request: web.Request):
        """ Authenticate, refresh or return the user credentials."""
        try:
            user, pwd = await self.get_payload(request)
        except Exception as err:
            pass
        if user and pwd:
            Default_SCOPE = ["User.ReadBasic.All"]
            # will use User/Pass Authentication
            app = self.get_msal_client()
            # Firstly, check the cache to see if this end user has signed in before
            accounts = app.get_accounts(username=user)
            result = None
            if accounts:
                result = app.acquire_token_silent(Default_SCOPE, account=accounts[0])
            if not result:
                # we need to get a new token from AAD
                result = app.acquire_token_by_username_password(
                    user, pwd, Default_SCOPE
                )
                try:
                    if 'access_token' in result:
                        access_token = result['access_token']
                        data = await self.get(url=self.userinfo_uri, token=access_token, token_type='Bearer')
                        userdata, uid = self.build_user_info(data)
                        # also, user information:
                        data = await self.create_user(request, uid, userdata, access_token)
                        # Redirect User to HOME
                        return self.home_redirect(request, token=data["token"], token_type='Bearer')
                    else:
                        if 65001 in result.get('error_codes', []):
                            # AAD requires user consent for U/P flow
                            print(
                                "Visit this to consent:", app.get_authorization_request_url(Default_SCOPE)
                            )
                        else:
                            error = result.get('error')
                            desc = result.get('error_description')
                            correlation = result.get("correlation_id")
                            message = f"Azure {error}: {desc}, correlation id: {correlation}"
                            logging.exception(message)
                            raise web.HTTPForbidden(
                                reason=message
                            )
                except Exception as err:
                    raise web.HTTPForbidden(
                        reason=f"Azure: Invalid Response from Server {err}."
                    )
        else:
            domain_url = self.get_domain(request)        
            self.redirect_uri = self.redirect_uri.format(domain=domain_url, service=self._service_name)
            SCOPE = ["https://graph.microsoft.com/.default"]
            app = self.get_msal_app()
            try:
                flow = app.initiate_auth_code_flow(
                    scopes=SCOPE,
                    redirect_uri=self.redirect_uri,
                    domain_hint=AZURE_ADFS_DOMAIN,
                    max_age=120
                )
                async with await request.app['redis'].acquire() as redis:
                    state = flow['state']
                    await redis.setex(f'azure_auth_{state}', json.dumps(flow), timeout=120)
                login_url = flow['auth_uri']
                return self.redirect(login_url)
            except Exception as err:
                raise NavException(
                    f"Azure: Client doesn't have info for Authentication: {err}"
                )


    async def auth_callback(self, request: web.Request):
        try:
            try:
                redis = request.app['redis']
            except Exception as err:
                raise Exception(
                    f'Azure Auth: Cannot get a Redis Cache connection: {err}'
                )
            auth_response = dict(request.rel_url.query.items())
            state = None
            SCOPE = ["https://graph.microsoft.com/.default"]
            try:
                state = auth_response['state']
            except Exception as err:
                raise Exception('Azure: Wrong authentication Callback, missing State.')
            try:
                async with await request.app['redis'].acquire() as redis:
                    result = await redis.get(f'azure_auth_{state}')
                    flow = json.loads(result)
            except Exception as err:
                raise Exception(f'Azure: Error reading Flow State from Cache: {err}')
            app = self.get_msal_app()
            try:
                result = app.acquire_token_by_auth_code_flow(
                    auth_code_flow=flow,
                    auth_response=auth_response
                )
                if 'token_type' in result:
                    token_type = result['token_type']
                    expires_in = result['expires_in']
                    access_token = result['access_token']
                    refresh_token = result['refresh_token']
                    id_token = result['id_token']
                    # logging.debug(f"Received access token: {id_token}")
                    # claims = jwt.decode(
                    #     id_token,
                    #     algorithms=['RS256', 'RS384', 'RS512'],
                    #     verify=False
                    # )
                    # print(f"JWT claims:\n {claims}")
                    claims = result['id_token_claims']
                    # getting user information:
                    try:
                        data = await self.get(url=self.userinfo_uri, token=access_token, token_type=token_type)
                        # build user information:
                        userdata, uid = self.build_user_info(data)
                        #userdata['access_token'] = access_token
                        userdata['id_token'] = id_token
                        #userdata['refresh_token'] = refresh_token
                        userdata['claims'] = claims
                        data = await self.create_user(request, uid, userdata, access_token)
                    except Exception as err:
                        logging.exception('Azure: Error getting User information')
                        raise web.HTTPForbidden(
                            reason=f"Azure: Error with User Information: {err}"
                        )
                    # Redirect User to HOME
                    return self.home_redirect(request, token=data['token'], token_type=token_type)
                elif 'error' in result:
                    error = result['error']
                    desc = result['error_description']
                    message = f"Azure {error}: {desc}"
                    logging.exception(message)
                    response = {
                        "message": message
                    }
                    raise web.HTTPForbidden(
                        reason=message
                    )
                else:
                    raise web.HTTPForbidden(
                        reason="Azure: Invalid Response from Server."
                    )
            except Exception as err:
                logging.exception(err)
                return self.redirect(uri=self.login_failed_uri)
        except Exception as err:
            logging.exception(err)
            return self.redirect(uri=self.login_failed_uri)

    async def logout(self, request):
        pass

    async def finish_logout(self, request):
        pass

    async def check_credentials(self, request):
        """ Authentication and create a session."""
        return True



# -*- coding: utf-8 -*-
import sys
import base64
import json
import importlib
from pathlib import Path
from types import ModuleType
from typing import (
    Any,
    Dict,
    List,
    Tuple
)
from cryptography import fernet
# Import Config Class
from navconfig import (
    BASE_DIR,
    config,
    DEBUG
)
from navconfig.logging import (
    logging
)

"""
Routes
"""
APP_NAME = config.get('APP_NAME', fallback='Navigator')
APP_DIR = BASE_DIR.joinpath("apps")

logging.debug(f'::: STARTING APP: {APP_NAME} in path: {APP_DIR} ::: ')

APP_HOST = config.get('APP_HOST', fallback='0.0.0.0')
APP_PORT = config.get('APP_PORT', fallback=5000)
APP_URL = config.get('APP_URL', fallback=f"http://localhost:{APP_PORT}")

TEMP_DIR = config.get("TEMP_DIR", fallback="/tmp")
NAV_DIR = BASE_DIR.joinpath("navigator")
TEMPLATE_DIR = BASE_DIR.joinpath("templates")
SERVICES_DIR = BASE_DIR.joinpath("services")
HOSTS = [e.strip() for e in list(config.get("HOSTS", fallback="localhost").split(","))]
DOMAIN = config.get("DOMAIN", fallback="dev.local")
STATIC_DIR = config.get('STATIC_DIR', fallback='static/')
# Temp File Path
files_path = BASE_DIR.joinpath("temp")

"""
Security and debugging
"""
# SECURITY WARNING: keep the secret key used in production secret!
fernet_key = fernet.Fernet.generate_key()
new_secret = base64.urlsafe_b64decode(fernet_key)
SECRET_KEY = config.get("SECRET_KEY", fallback=new_secret)

# used by tokenauth with RNC.
PARTNER_KEY = config.get("PARTNER_KEY")
CYPHER_TYPE = config.get("CYPHER_TYPE", fallback="RNC")

"""
Development
"""
# Debug
#
DEBUG = config.getboolean("DEBUG", fallback=False)
PRODUCTION = bool(config.getboolean("PRODUCTION", fallback=(not DEBUG)))
LOCAL_DEVELOPMENT = DEBUG == True and sys.argv[0] == "run.py"
USE_SSL = config.getboolean("ssl", "SSL", fallback=False)

"""
Timezone
"""
# Timezone (For parsedate)
# https://dateparser.readthedocs.io/en/latest/#timezone-and-utc-offset
TIMEZONE = config.get("TIMEZONE", fallback="UTC")

"""
Environment
"""
if DEBUG and LOCAL_DEVELOPMENT:
    ENV = "development"
    CSRF_ENABLED = False
    SSL = False
    SSL_VERIFY = False
    SSL_CERT = None
    SSL_KEY = None
    PREFERRED_URL_SCHEME = "http"
    ENABLE_TOKEN_APP = False
else:
    if PRODUCTION == False and DEBUG == True:
        ENV = "development"
        CSRF_ENABLED = False
    elif PRODUCTION == True and DEBUG == True:
        ENV = "staging"
        CSRF_ENABLED = True
    else:
        ENV = "production"
        CSRF_ENABLED = True
    try:
        SSL_CERT = config.get("CERT")
        SSL_KEY = config.get("KEY")
        PREFERRED_URL_SCHEME = "https"
    except Exception as e:
        SSL_CERT = None
        SSL_KEY = None
        PREFERRED_URL_SCHEME = "http"

"""
Basic Information
"""
EMAIL_CONTACT = config.get("EMAIL_CONTACT", section="info", fallback="foo@example.com")
API_NAME = config.get("API_NAME", section="info", fallback="Navigator")

#######################
##
## APPS CONFIGURATION
##
#######################
"""
Main Database
"""
PG_USER = config.get("DBUSER")
PG_HOST = config.get("DBHOST", fallback="localhost")
PG_PWD = config.get("DBPWD")
PG_DATABASE = config.get("DBNAME", fallback="navigator")
PG_PORT = config.get("DBPORT", fallback=5432)

asyncpg_url = "postgres://{user}:{password}@{host}:{port}/{db}".format(
    user=PG_USER, password=PG_PWD, host=PG_HOST, port=PG_PORT, db=PG_DATABASE
)
default_dsn = asyncpg_url

"""
Auth and Cache
"""

"""
REDIS SESSIONS
"""
CACHE_HOST = config.get("CACHEHOST", fallback="localhost")
CACHE_PORT = config.get("CACHEPORT", fallback=6379)
CACHE_URL = "redis://{}:{}".format(CACHE_HOST, CACHE_PORT)
REDIS_SESSION_DB = config.get("REDIS_SESSION_DB", fallback=0)
CACHE_PREFIX = config.get('CACHE_PREFIX', fallback='navigator')

"""
Authentication System
"""
AUTHENTICATION_BACKENDS = (
)

AUTHORIZATION_BACKENDS = [
    e.strip()
    for e in list(
        config.get("AUTHORIZATION_BACKENDS", fallback="allow_hosts").split(",")
    )
]

AUTHORIZATION_MIDDLEWARES = (
)


# Basic Authentication
AUTH_TOKEN_ISSUER = config.get('AUTH_TOKEN_ISSUER', fallback='Navigator')
AUTH_PWD_DIGEST = config.get("AUTH_PWD_DIGEST", fallback="sha256")
AUTH_PWD_ALGORITHM = config.get("AUTH_PWD_ALGORITHM", fallback="pbkdf2_sha256")
AUTH_PWD_LENGTH = config.get("AUTH_PWD_LENGTH", fallback=32)
AUTH_PWD_SALT_LENGTH = config.get("AUTH_PWD_SALT_LENGTH", fallback=6)
AUTH_USERNAME_ATTRIBUTE = config.get(
    'AUTH_USERNAME_ATTRIBUTE', fallback='username'
)
CREDENTIALS_REQUIRED = config.getboolean(
    "AUTH_CREDENTIALS_REQUIRED", fallback=False
)
AUTH_USER_MODEL = config.get(
    "AUTH_USER_MODEL", fallback="navigator.auth.models.User"
)
AUTH_GROUP_MODEL = config.get(
    "AUTH_GROUP_MODEL", fallback="navigator.auth.models.Group"
)
AUTH_SESSION_OBJECT = config.get(
    "AUTH_SESSION_OBJECT", fallback="session"
)
AUTH_REDIRECT_URI = config.get('AUTH_REDIRECT_URI', section="auth", fallback=APP_URL)
AUTH_LOGIN_FAILED_URI = config.get('AUTH_REDIRECT_URI', section="auth", fallback=APP_URL)

DEFAULT_MAPPING = {
    "user_id": "user_id",
    "username": "username",
    "password": "password",
    "first_name": "first_name",
    "last_name": "last_name",
    "email": "email",
    "enabled": "is_active",
    "superuser": "is_superuser",
    "last_login": "last_login",
    "title": "title",
}
mapping = config.get('AUTH_USER_MAPPING')
if mapping:
    USER_MAPPING = json.loads(mapping)
else:
    USER_MAPPING = DEFAULT_MAPPING

USERS_TABLE = config.get("AUTH_USERS_TABLE", fallback="vw_users")

ALLOWED_HOSTS = [
    e.strip()
    for e in list(config.get("ALLOWED_HOSTS", section="auth", fallback="localhost*").split(","))
]

"""
Session Storage
"""
SESSION_NAME = "{}_SESSION".format(
    config.get("APP_TITLE", fallback="NAVIGATOR").upper()
)
JWT_ALGORITHM = config.get("JWT_ALGORITHM", fallback="HS256")
SESSION_PREFIX = '{}_session'.format(CACHE_PREFIX)
SESSION_TIMEOUT = config.getint('SESSION_TIMEOUT', fallback=360000)
SESSION_KEY = config.get('SESSION_KEY', fallback='id')
SESSION_STORAGE = 'NAVIGATOR_SESSION_STORAGE'
SESSION_OBJECT = 'NAV_SESSION'
SESSION_URL = f"redis://{CACHE_HOST}:{CACHE_PORT}/{REDIS_SESSION_DB}"
SESSION_USER_PROPERTY = config.get('SESSION_USER_PROPERTY', fallback='user')

"""
 Memcache
"""
MEMCACHE_HOST = config.get("MEMCACHE_HOST", "localhost")
MEMCACHE_PORT = config.get("MEMCACHE_PORT", 11211)

# get configuration settings (user can override settings).
try:
    from navconfig.conf import *
except ImportError:
    from settings.settings import *

"""
Final: Config dict for aiohttp
"""
Context = {
    "DEBUG": DEBUG,
    "DEVELOPMENT": (not PRODUCTION),
    "LOCAL_DEVELOPMENT": LOCAL_DEVELOPMENT,
    "PRODUCTION": PRODUCTION,
    "SECRET_KEY": SECRET_KEY,
    "env": ENV,
    "cache_url": CACHE_URL,
    "asyncpg_url": asyncpg_url,
    "default_dsn": default_dsn,
}

"""
Applications
"""
INSTALLED_APPS: List = []
DATABASES: Dict = {}

if APP_DIR.is_dir():
    for item in APP_DIR.iterdir():
        if item.name != "__pycache__":
            if item.is_dir():
                name = item.name
                if not name in INSTALLED_APPS:
                    app_name = "apps.{}".format(item.name)
                    path = APP_DIR.joinpath(name)
                    url_file = path.joinpath("urls.py")
                    try:
                        i = importlib.import_module(app_name, package="apps")
                        if isinstance(i, ModuleType):
                            # is a Navigator Program
                            INSTALLED_APPS += (app_name,)
                    except ImportError as err:
                        print("ERROR: ", err)
                        continue
                    # schema configuration
                    DATABASES[item.name] = {
                        "NAME": PG_DATABASE,
                        "USER": PG_USER,
                        "OPTIONS": {
                            "options": "-c search_path=" + item.name + ",public",
                        },
                        "SCHEMA": item.name,
                        "PASSWORD": PG_PWD,
                        "HOST": PG_HOST,
                        "PORT": PG_PORT,
                    }

Context["DATABASES"] = DATABASES

"""
Per-Program Settings
"""
program: str
for program in INSTALLED_APPS:
    settings = "apps.{}.settings".format(program)
    try:
        i = importlib.import_module(settings)
        globals()[program] = i
    except ImportError as err:
        pass



import asyncio
from functools import wraps
from typing import Callable
from aiohttp import web
from navigator.auth.sessions import get_session
from navigator.auth import get_userdata

def auth_required(func: Callable) -> Callable:
    """Decorator to check if an user has been authenticated for this request.
    """
    @wraps(func)
    async def _wrap(*args, **kwargs):
        # handling functions and class methods
        handler = args[0]
        if isinstance(handler, web.Request):
            request = handler
        else:
            request = (args[-1].request if isinstance(args[-1], web.View)
                       else args[-1])
        if request is None:
            raise ValueError(f'web.Request was not found in arguments. {fn!s}')
        app = request.app
        auth = app["auth"]
        for backend in auth.backends():
            if await backend.check_credentials(request):
                return await func(*args, **kwargs)
            else:
                continue
        raise web.HTTPUnauthorized(
            reason="Access Denied",
            headers={
                hdrs.CONTENT_TYPE: 'text/html; charset=utf-8',
                hdrs.CONNECTION: 'keep-alive',
            }
        )
        # router = app.router
        # return web.HTTPFound(router["login"].url_for())
    return _wrap

def has_permission(func, permission, context=None):
    """Check user permissions.
    Return Function if the identity is allowed the permission in the
    current context, else raise ``web.HTTPUnauthorized()``
    """
    @wraps(func)
    async def _wrap(*args, **kwargs):
        pass
    return _wrap

def restricted(func, restrictions):
    """Restrict the Handler to certain Groups/Restriction Classes.
    Return Function if the identity is allowed under current restrictions,
    else raise ``web.HTTPUnauthorized()``
    """
    @wraps(func)
    async def _wrap(*args, **kwargs):
        pass
    return _wrap

def login_required(func: Callable) -> Callable:
    func.__login_required__ = True
    return func



"""Nav Middleware.

Navigator Authorization Middlewares.
"""

from .troc import troctoken_middleware
from .jwt import jwt_middleware
from .django import django_middleware
from .token import token_middleware
from .apikey import apikey_middleware

__all__ = [
    "troctoken_middleware",
    "jwt_middleware",
    "django_middleware",
    "token_middleware",
    "apikey_middleware"
]
